//------------------------------------------------------------------------------
/// @brief ARM - RTE communication library
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription rte_comm.c
/// ARM - RTE communication library implementation
/// 
//------------------------------------------------------------------------------

#include "rte_comm.h"

static size_t PipeRead(char* filename, char* buf, size_t size) {
  FILE* fp = popen(filename, "r");
  size_t read_size = fread(buf, sizeof(char), size, fp);
  pclose(fp);
  return read_size;
}

static int GetNumCores(void)
{
  char buf[8];
  PipeRead("/bin/cat /sys/devices/system/cpu/possible", buf, sizeof(buf) - 1);
  int from, to;
  sscanf(buf, "%d-%d", &from, &to);
  int detected_cores = to - from + 1;

  return detected_cores;
}

int __libc_start_main(int (*main) (int, char **, char **),
    int argc,
    char ** ubp_av,
    void (*init) (void),
    void (*fini) (void),
    void (*rtld_fini) (void),
    void (*stack_end))
{
  static void (*main_p) (int (*main) (int, char **, char **),
      int argc,
      char ** ubp_av,
      void (*init) (void),
      void (*fini) (void),
      void (*rtld_fini) (void),
      void (*stack_end));
  char* error;

  if (RTE_StartApplication(ubp_av[0])) {
    fprintf(stderr, "RTE StartApplication error\n");
  }

  if (!main_p) {
    main_p = dlsym(RTLD_NEXT, "__libc_start_main");
    if ((error = dlerror()) != NULL) {
      fputs(error, stderr);
      exit(1);
    }
  }

  main_p(main, argc, ubp_av, init, fini, rtld_fini, stack_end);
}

void exit(int status)
{
  static void (*exit_p) (int status);
  char* error;

  if (RTE_FinishApplication()) {
    fprintf(stderr, "RTE FinishApplication error\n");
  }

  if (!exit_p) {
    exit_p = dlsym(RTLD_NEXT, "exit");
    if ((error = dlerror()) != NULL) {
      fputs(error, stderr);
      exit(1);
    }
  }

  exit_p(status);
}

void InitSHM(void)
{
  key = (key_t)getpid();

  if ((shmid = shmget(key, sizeof(app_context) + num_cores * sizeof(int),
          IPC_CREAT|0666)) == -1) {
    perror("shmget error");
    exit(-1);
  }

  if ((shmaddr = shmat(shmid, NULL, 0)) == (void*)-1) {
    perror("shmat error");
    exit(-1);
  }

  shm_opened = 1;

  printf("key: %d ", key);
  printf("shmid: %d ", shmid);
  printf("shmaddr: %lu ", (unsigned long)shmaddr);
  printf("shm_opened: %d\n", shm_opened);

  if (shm_opened) {
    app_context context;
    int i = 0;

    context.region_id = 0;
    strncpy(context.kernel_name, "unknown", 7);

    if (!memcpy((void*)(unsigned long)shmaddr, &context, sizeof(app_context))) {
      perror("memcpy error");
      exit(-1);
    }
  }
}

void CleanUpSHM(void)
{
  if (shm_opened) {
    if (shmctl(shmid, IPC_RMID, 0) == -1) {
      perror("shmdt error");
      exit(-1);
    }

    shm_opened = 0;
  }
}

int RTE_StartApplication(char* buf)
{
  app_begin_data data;
  int command = APP_BEGIN;
  struct sockaddr_in addr;
  int rte_socket;
  int ack = 0;

  //num_cores = GetNumCores();
  num_cores = numa_num_task_cpus();
  printf("detected %d cores\n", num_cores);

  //num_cores = 72;

#if defined(DEBUG_MODE)
  printf("RTE_StartApplication %s(%d)\n", buf, getpid());
#endif

  InitSHM();

  /* initialize application runtime */
  /* register its pid and application name */
  data.pid = getpid();
  strncpy(data.name, buf, strlen(buf));

  rte_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (rte_socket == -1) {
    perror("create rte socket error");
    return -1;
  }

  // connection with rte monitor process on localhost
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(11111);
  addr.sin_addr.s_addr = inet_addr("127.0.0.1");

  if (connect(rte_socket, (struct sockaddr*)&addr, sizeof(addr))) {
    perror("connection error");
    return -1;
  }

  if (write(rte_socket, &command, sizeof(int)) <= 0) {
    perror("send app command 1 error");
    return -1;
  }

  if (write(rte_socket, &data, sizeof(app_begin_data)) <= 0) {
    perror("send app msg 1 error");
    return -1;
  }

  if (read(rte_socket, &ack, sizeof(int)) <= 0) {
    perror("receive ack error: socket error");
    return -1;
  }

  if (ack != 1) {
    perror("receive ack error: unexpected ack value");
    return -1;
  }

  close(rte_socket);

  return 0;
}

int RTE_FinishApplication(void)
{
  int command = APP_END;
  app_end_data data;
  struct sockaddr_in addr;
  int rte_socket;
  int ack = 0;

#if defined(DEBUG_MODE)
  printf("RTE_FinishApplication (%d)\n", getpid());
#endif

  data.pid = getpid();

  rte_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (rte_socket == -1) {
    perror("create rte socket error");
    return -1;
  }

  // connection with rte monitor process on localhost
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(11111);
  addr.sin_addr.s_addr = inet_addr("127.0.0.1");

  if (connect(rte_socket, (struct sockaddr*)&addr, sizeof(addr))) {
    perror("connection error");
    return -1;
  }

  if (write(rte_socket, &command, sizeof(int)) <= 0) {
    perror("send app command 4 error");
    return -1;
  }

  if (write(rte_socket, &data, sizeof(app_end_data)) <= 0) {
    perror("send app msg 1 error");
    return -1;
  }

  if (read(rte_socket, &ack, sizeof(int)) <= 0) {
    perror("receive ack error: socket error");
    return -1;
  }

  if (ack != 1) {
    perror("receive ack error: unexpected ack value");
    return -1;
  }

  close(rte_socket);

  CleanUpSHM();

  return 0;
}

