//------------------------------------------------------------------------------
/// @brief ARM-RTE application description
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription app.cpp
/// Basic implementation of the applications on RTE.
/// It contains information of application profiles and application states
///
//------------------------------------------------------------------------------

#include "app.h"
#include "scheduler.h"
#include <math.h>
#include <string.h>
#include <sched.h>
#include <stdio.h>

//------------------------------------------------------------------------------
RegionProfiles::RegionProfiles(int region_no, char* kernel_name)
    : id_(region_no)
{
    name_ = (char*)malloc(strlen(kernel_name) * sizeof(char));
    if (!strncpy(name_, kernel_name, strlen(kernel_name))) {
        perror("strcpy error");
        exit(-1);
    }

    llc_miss_rate_ = 1.0;
    sum_llc_misses_ = 0;
    sum_llc_total_cycles_ = 0;

    dram_req_rate_ = 1.0;
    sum_dram_reqs_ = 0;
    sum_dram_total_cycles_ = 0;

    is_profiled_ = 0;
}

RegionProfiles::~RegionProfiles()
{
}

void RegionProfiles::UpdateLLCAccessRate(unsigned long llc_accesses,
        unsigned long total_cycles)
{
    //sum_llc_accesses_ += llc_accesses;
    //sum_llc_access_total_cycles_ += total_cycles;
    //llc_access_rate_ = (double)sum_llc_accesses_/(double)sum_llc_access_total_cycles_;
    llc_access_rate_ = (double)llc_accesses / (double)total_cycles;
    std::cout << "UpdateLLCAccessRate: " << llc_access_rate_
        << " region_id: " << id_ << std::endl;
}

void RegionProfiles::UpdateLLCMissRate(unsigned long llc_misses,
        unsigned long total_cycles)
{
    //sum_llc_misses_ += llc_misses;
    //sum_llc_total_cycles_ += total_cycles;
    //llc_miss_rate_ = (double)sum_llc_misses_/(double)sum_llc_total_cycles_;
    llc_miss_rate_ = (double)llc_misses / (double)total_cycles;
    std::cout << "UpdateLLCMissRate: " << llc_miss_rate_
        << " region_id: " << id_ << std::endl;
}

void RegionProfiles::UpdateLLCMissAccessRate(unsigned long llc_misses,
        unsigned long llc_accesses)
{
    llc_miss_access_rate_ = (double)llc_misses / (double)llc_accesses;
    //llc_miss_access_rate_ = (double)sum_llc_misses_/(double)sum_llc_accesses_;
    std::cout << "UpdateLLCMissAccessRate: " << llc_miss_access_rate_
        << " region_id: " << id_ << std::endl;
}

void RegionProfiles::UpdateDRAMReqRate(unsigned long dram_reqs,
        unsigned long total_cycles)
{
    //sum_dram_reqs_ += dram_reqs;
    //sum_dram_total_cycles_ += total_cycles;
    //dram_req_rate_ = (double)sum_dram_reqs_ / (double)sum_dram_total_cycles_;
    dram_req_rate_ = (double)dram_reqs / (double)total_cycles;
    std::cout << "UpdateDRAMReqRate: " << dram_req_rate_
        << " region_id: " << id_ << std::endl;
}

void RegionProfiles::DumpPerformanceProfiles(ostream & dout)
{
    dout << endl << endl;
    dout << "parallel region number," << id() << endl;
    dout << "parallel region name," << name() << endl;
}

//------------------------------------------------------------------------------
AddressSpace::AddressSpace(int offset, int size)
    : offset_(offset), size_(size)
{
}

AddressSpace::~AddressSpace()
{
}

PA AddressSpace::CPAtoPA(CPA cpa)
{
    PA pa = cpa + offset_;

    return pa;
}

CPA AddressSpace::PAtoCPA(PA pa)
{
    CPA cpa = pa - offset_;

    if (cpa >= 0 && cpa < size_) {
        return cpa;
    } else {
        cout << "invalid CPA address: " << cpa << endl;
        return -1;
    }
}

void AddressSpace::AddressSpaceInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "----------------------------------------"
        << "----------------------------------------" << endl;
    out << "address space description" << endl;
    out << "system memory offset: " << offset_
        << " memory size: " << size_ << endl;

    out.close();

    return;
}

//------------------------------------------------------------------------------
ParApp::ParApp(pid_t pid, string name,
        CoreAllocator* scheduler, CPUArchitecture* architecture)
    : pid_(pid), name_(name)
{
    // get the scheduler in which this application runs
    assert(scheduler != NULL);
    scheduler_ = scheduler;

    // get the target architecture where this application runs
    assert(architecture != NULL);
    arch_ = architecture;

    // initialize application variables
    allocated_tiles_ = 0;  // currently allocated tiles
    reserved_tiles_ = 0;
    current_region_  = 0;  // application starting point: sequential region
    cached_region_ = 0;
    curr_cpu_cycles_ = 0;

    need_profiling_ = false;

    // initialize shared memory area
    InitSHM();
}

ParApp::~ParApp()
{
    //ReleaseSHM();
}

double ParApp::GetUtil()
{
    double sum_cpu_util = 0;
    size_t num_alloc_tiles = allocated_tiles_;

#if DEBUG_MODE
    std::cout << "App (" << pid() << ") " << std::endl;
#endif

    /* sum the CPU utilization of every tile allocated by this app */
    for (list<CPUTile*>::iterator tile_it = allocated_tile_list_.begin();
                                    tile_it != allocated_tile_list_.end();
                                    tile_it++)
    {
        CPUTile* tile = *tile_it;
        //double tile_cpu_util = tile->GetLLCMissRate();
        double tile_cpu_util = tile->GetCPUUtilization();
        sum_cpu_util += tile_cpu_util;
#if DEBUG_MODE
        std::cout << "\tTile " << tile->id()
            << " Util: " << tile_cpu_util << std::endl;
#endif
    }

    /* compute the avg CPU utilization among all tiles allocated by this app */
    double cpu_util = sum_cpu_util / num_alloc_tiles;
#if DEBUG_MODE
    std::cout << std::endl << "\tOverall CPU Util: " << cpu_util << std::endl;
#endif

    return cpu_util;
}

rte_cpu_set ParApp::GetRTECPUSet(void)
{
    rte_cpu_set cpuset;
    RTE_CPU_ZERO(&cpuset);

    for (list<CPUTile*>::iterator it = tile_list_.begin();
            it != tile_list_.end();
            it++) {
        CPUTile* tile = (*it);
        RTE_CPU_SET(tile->id(), &cpuset);
    }

    return cpuset;
}

void ParApp::DumpPerformanceProfiles(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << endl;
    out << "Application,Performance,Profiles" << endl;
    out << name_ << endl;

    out << "application name,application pid,"
        << "target architecture,architecture model" << endl
        << name_ << "," << pid_ << "," << arch_->arch_type() << ","
        << arch_->vendor() << endl;

    for (map<int, RegionProfiles*>::iterator it = profiles_.begin();
            it != profiles_.end();
            it++) {
        RegionProfiles* profiles = it->second;
        profiles->DumpPerformanceProfiles(out);
    }

    out << endl;
    out.close();

    return;
}

void ParApp::ShowAppInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "----------------------------------------"
        << "----------------------------------------" << endl;
    out << "application name: " << name_ << " pid: " << pid_ << endl;
    out << "target architecture: " << arch_->arch_type() << endl;
    out << "architecture model: " << arch_->vendor() << endl;

    out.close();

    return;
}

void ParApp::ScheduleCPUs(void)
{
    std::cout << "app: " << pid_ << " allocated: ";
    if (shm_opened_) {
        for (int i = 0; i < arch_->num_cores(); i++)
            alloc_map_[i] = 0;

        while (!allocated_tile_list_.empty()) {
            CPUTile* tile = allocated_tile_list_.front();

            list<CPUTile*>::iterator it = allocated_tile_list_.begin();
            while (it != allocated_tile_list_.end()) {
                if (tile == (*it)) {
                    tile_list_.erase(it);
                    break;
                }
                it++;
            }
        }

        /* set CPU affinity for the master thread */
        cpu_set_t cpuset;
        CPU_ZERO(&cpuset);

        list<CPUTile*> tile_list = GetTileSet();
        for (list<CPUTile*>::iterator it = tile_list.begin();
             it != tile_list.end();
             it++) {
            CPUTile* tile = *it;
            std::cout << tile->id() << " ";

            for (int j = 0; j < tile->num_cores(); j++) {
                CPUCore* core = tile->GetCore(j);
                int node_num = core->GetCPUNode()->id();
                int core_num = core->GetNodeIndex();
                node_alloc_map_[node_num][core_num] = pid();

                //std::cout << "[sched_setaffinity] pid: " << pid() << " core_id: " << core->id() << " tile_id: " << tile->id() << std::endl;
                CPU_SET(core->id(), &cpuset);
            }

            allocated_tile_list_.push_back(tile);
        }

        if (tile_list.size()) {
            if (sched_setaffinity((pid_t)pid(), sizeof(cpu_set_t), &cpuset)) {
                perror("sched_setaffinity error");
                exit(-1);
            }
        }
    }
    std::cout << std::endl;

    return;

    if (shm_opened_) {
        list<CPUTile*> tile_list = GetTileSet();
        int* available = (int*)calloc(arch_->num_tiles(), sizeof(int));

        for (list<CPUTile*>::iterator it = tile_list.begin();
                it != tile_list.end();
                it++) {
            CPUTile* tile = (*it);
            available[tile->id()] = pid();
        }

        /* don't bother delegate threads */
        std::list<int> delegated_list = scheduler_->GetDelegatedList();
        for (std::list<int>::iterator it = delegated_list.begin();
                it != delegated_list.end();
                it++) {
            int delegated = *it;
            available[delegated] = 0;
            //cout << delegated << " ";
        }
        //cout << "are delegated" << endl;

        /* reserve my delegate thread */
        int delegate = GetDelegate();
        available[delegate] = pid();
        if (!memcpy((void*)((unsigned long)shmaddr_ + sizeof(int)),
                    &delegate, sizeof(int))) {
            perror("memcpy error");
            exit(-1);
        }

        //if (available[0] != pid())
        //  available[last_tile] = 0;

        //if (available[0] != pid()) {
        //  /* set CPU affinity for the master thread */
        //  cpu_set_t cpuset;
        //  CPU_ZERO(&cpuset);

        //  available[last_tile] = 0;
        //  CPU_SET(last_tile, &cpuset);

        //  if (sched_setaffinity(pid(), sizeof(cpu_set_t), &cpuset)) {
        //    perror("sched_setaffinity error");
        //    exit(-1);
        //  }
        //} else {
        //  /* set CPU affinity for the master thread */
        //  cpu_set_t cpuset;
        //  CPU_ZERO(&cpuset);
        //  CPU_SET(0, &cpuset);

        //  if (sched_setaffinity(pid(), sizeof(cpu_set_t), &cpuset)) {
        //    perror("sched_setaffinity error");
        //    exit(-1);
        //  }
        //}

        if (!memcpy((void*)((unsigned long)shmaddr_ + 2*sizeof(int)),
                    available, arch_->num_tiles() * sizeof(int))) {
            perror("memcpy error");
            exit(-1);
        }
    }
}

uint64_t ParApp::GetCurrCpuCycles()
{
    return curr_cpu_cycles_;
}

double ParApp::GetLLCMissRate(int tiles)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    if (it == profiles_.end())
        return 0.0;

    RegionProfiles* profiles = it->second;
    return profiles->llc_miss_rate_ * (double)tiles;
}

double ParApp::GetDRAMReqRate(int tiles)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    if (it == profiles_.end())
        return 0.0;

    RegionProfiles* profiles = it->second;
    return profiles->dram_req_rate_ * (double)tiles;
}

void ParApp::UpdateAppContext(void)
{
    if (shm_opened_) {
        app_context context;

        /* communicate with application runtime */
        if (!memcpy(&context, shmaddr_, sizeof(app_context))) {
            perror("memcpy error");
            exit(-1);
        }

        current_region_ = context.region_id;
        memset(current_kernel_name_, 0x00, 4096);
        if (!strncpy(current_kernel_name_,
                    context.kernel_name,
                    strlen(context.kernel_name))) {
            perror("strcpy error");
            exit(-1);
        }
    }
}

void ParApp::UpdateLLCAccessRate(unsigned long llc_accesses,
        unsigned long total_cycles)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    std::cout << "app: " << pid_ << " name: " << name_
        << " current region: " << current_region_
        << " UpdateLLCAccessRate" << std::endl;

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;
        profiles->UpdateLLCAccessRate(llc_accesses, total_cycles);
    } else {
        RegionProfiles* profiles = new RegionProfiles(current_region_,
              current_kernel_name_);
        profiles_.insert(pair<int, RegionProfiles*>(current_region_, profiles));
        profiles->UpdateLLCAccessRate(llc_accesses, total_cycles);
    }
}

void ParApp::UpdateLLCMissRate(unsigned long llc_misses,
        unsigned long total_cycles)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    std::cout << "app: " << pid_ << " name: " << name_
        << " current region: " << current_region_
        << " UpdateLLCMissRate" << std::endl;

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;
        profiles->UpdateLLCMissRate(llc_misses, total_cycles);
    } else {
        RegionProfiles* profiles = new RegionProfiles(current_region_,
                current_kernel_name_);
        profiles_.insert(pair<int, RegionProfiles*>(current_region_, profiles));
        profiles->UpdateLLCMissRate(llc_misses, total_cycles);
    }
}

void ParApp::UpdateDRAMReqRate(unsigned long dram_reqs, unsigned long total_cycles)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    std::cout << "app: " << pid_ << " name: " << name_
        << " current region: " << current_region_
        << " UpdateDRAMReqRate" << std::endl;

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;
        profiles->UpdateDRAMReqRate(dram_reqs, total_cycles);
    } else {
        RegionProfiles* profiles = new RegionProfiles(current_region_,
                current_kernel_name_);
        profiles_.insert(pair<int, RegionProfiles*>(current_region_, profiles));
        profiles->UpdateDRAMReqRate(dram_reqs, total_cycles);
    }
}

void ParApp::UpdateCurrCpuCycles(unsigned long total_cycles)
{
    curr_cpu_cycles_ = total_cycles;
}

void ParApp::UpdateLLCMissAccessRate(unsigned long llc_misses,
        unsigned long llc_accesses)
{
    map<int, RegionProfiles*>::iterator it = profiles_.find(current_region_);

    std::cout << "app: " << pid_ << " name: " << name_
        << " current region: " << current_region_
        << " UpdateLLCMissAccessRate" << std::endl;

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;
        profiles->UpdateLLCMissAccessRate(llc_misses, llc_accesses);
    } else {
        RegionProfiles* profiles = new RegionProfiles(current_region_,
                current_kernel_name_);
        profiles_.insert(pair<int, RegionProfiles*>(current_region_, profiles));
        profiles->UpdateLLCMissAccessRate(llc_misses, llc_accesses);
    }
}

double ParApp::GetLLCMissAccessRate()
{
    map<int, RegionProfiles*>::iterator it =
        profiles_.find(current_region_);

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;
        return profiles->GetLLCMissAccessRate();
    }

    return 0.0;
}

void ParApp::SetRegionProfiled()
{
    map<int, RegionProfiles*>::iterator it =
        profiles_.find(current_region_);

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;

        profiles->is_profiled_++;
    }
}

bool ParApp::IsRegionProfiled()
{
    map<int, RegionProfiles*>::iterator it =
        profiles_.find(current_region_);

    if (it != profiles_.end()) {
        RegionProfiles* profiles = it->second;

        if (profiles->is_profiled_) return true;

        //if (profiles->is_profiled_ > 1)
        //    return true;
    }
    return false;
}

void ParApp::ShowAppResource(ostream & dout)
{
    // core allocation information
    dout << "----------------------------------------"
        << "----------------------------------------" << endl;
    dout << "application name: " << name_ << " pid: " << pid_
        << " target architecture: " << arch_->vendor() << endl;
    dout << "application current region: " << current_region_ << endl;
    dout << "tiles: ";

    list<CPUTile*>::iterator it = tile_list_.begin();
    while (it != tile_list_.end()) {
        dout << "<" << (*it)->id() <<"> ";
        it++;
    }

    dout << endl;
}

void ParApp::AllocOneTile(CPUTile* tile)
{
    // add one tile
    tile_list_.push_back(tile);
    allocated_tiles_++;

    // mark application id to tile structure
    tile->RegisterApp(pid());
}

bool ParApp::DeAllocOneTile(CPUTile* tile)
{
    // delete one tile from the list
    list<CPUTile*>::iterator it = tile_list_.begin();
    while (it != tile_list_.end()) {
        if (tile == (*it)) {
            // delete it
            tile_list_.erase(it);
            allocated_tiles_--;

            // remove application id marking
            tile->CleanApp();
            return true;
        }

        it++;
    }

    return false;
}

bool ParApp::FlushAllTiles(void)
{
    while (!tile_list_.empty()) {
        CPUTile* tile = tile_list_.front();
        tile->CleanLLCExchangeFlag();
        DeAllocOneTile(tile);
    }

    return true;
}

void ParApp::InitSHM(void)
{
    key_ = pid_;

    size_t shm_size = sizeof(app_context) + arch_->num_cores()*sizeof(int);

    if ((shmid_ = shmget(pid_, shm_size, 0666)) < 0) {
        perror("shmget error");
        exit(-1);
    }

    if ((shmaddr_ = shmat(shmid_, NULL, 0)) == (void*)-1) {
        perror("shmat error");
        exit(-1);
    }

    vector<int> node_ids = arch_->node_ids();
    int num_nodes = arch_->num_cpu_nodes();

    alloc_map_ = (int*)((unsigned long)shmaddr_ + sizeof(app_context));

    node_alloc_map_ = (int**)malloc((node_ids.back() + 1) * sizeof(int*));

    unsigned long offset = 0;
    for (int i = 0; i < num_nodes; i++) {
        node_alloc_map_[node_ids[i]] = (int*)((unsigned long)alloc_map_ + offset);
        offset += arch_->GetCPUNode(i)->num_cores() * sizeof(int);
    }

    shm_opened_ = 1;
}

void ParApp::ReleaseSHM(void)
{
    if (shm_opened_) {
        if (shmdt(shmaddr_) == -1) {
            perror("shmdt error");
            exit(-1);
        }
    }
}

