//------------------------------------------------------------------------------
/// @brief ARM-RTE application description
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription app.h
/// This is the basic description of the applications on RTE.
/// it contains information of application profiles and application states
///
//------------------------------------------------------------------------------

#ifndef __APP_INCLUDED_
#define __APP_INCLUDED_

#include "arch.h"
#include "rte_cpu.h"

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <list>

#include "comm_interface.h"
#include "power_interface.h" // get_freq from GetUtil

#include <unistd.h>

using namespace std;

class CoreAllocator;

//------------------------------------------------------------------------------
/// @brief application parallel region profiles (obtained from online profiler)
///
/// application parallel region profiles
///
class RegionProfiles {
  friend class ParApp;
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    RegionProfiles(int region_no, char* kernel_name);

    /// @brief default destructor
    ~RegionProfiles();

    /// @}

    /// @name parallel region profile attributes
    /// @{

    int id(void) const { return id_; }
    const char* name() const { return name_; }

    double GetLLCMissAccessRate(void) {
        return llc_miss_access_rate_;
    }

    /// @}

  private:
    /// @name parallel region profile private attributes
    /// @{

    /// @brief update LLC miss rate
    ///
    /// @param LLC misses obtained from profiler
    /// @param elapsed cycles obtained from profiler
    void UpdateLLCAccessRate(unsigned long llc_accesses,
        unsigned long total_cycles);
    void UpdateLLCMissRate(unsigned long llc_misses,
        unsigned long total_cycles);
    void UpdateLLCMissAccessRate(unsigned long llc_misses,
        unsigned long llc_accesses);

    /// @brief update DRAM request rate
    ///
    /// @param DRAM requests obtained from profiler
    /// @param elapsed cycles obtained from profiler
    void UpdateDRAMReqRate(unsigned long dram_reqs, unsigned long total_cycles);

    /// @brief dump performance profiles of this parallel region to log stream
    void DumpPerformanceProfiles(ostream & dout);

    /// @}

    const int    id_;                 ///< parallel region number (id)
    char*        name_;               ///< kernel name

    double        llc_access_rate_;
    unsigned long sum_llc_accesses_;
    unsigned long sum_llc_access_total_cycles_;

    double        llc_miss_rate_;     ///< llc miss rate per tile
    unsigned long sum_llc_misses_;
    unsigned long sum_llc_total_cycles_;

    double        llc_miss_access_rate_;

    double        dram_req_rate_;     ///< darm request rate per tile
    unsigned long sum_dram_reqs_;
    unsigned long sum_dram_total_cycles_;

    int is_profiled_;
};

//------------------------------------------------------------------------------
/// @brief application address space
///
/// the first version of application address space
/// address space is abstracted with a offset value and size
///
class AddressSpace {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    ///
    /// @param offset system memory offset
    /// @param size   application memory size (i.e., pages)
    AddressSpace(int offset, int size);

    ~AddressSpace();

    /// @}

    /// @name address space attributes
    /// @{

    /// @brief get application memory offset
    PA GetOffset(void) { return offset_; }

    /// @brief get application memory offset
    int GetSize(void) { return size_; }

    /// @brief set application memory offset
    void SetOffset(PA offset) { offset_ = offset; }

    /// @brief set application memory size

    /// @brief translate from CPA to PA
    PA  CPAtoPA(CPA pfn);

    /// @brief translate from PA to CPA
    CPA PAtoCPA(PA pfn);

    /// @brief show address space brief information
    void AddressSpaceInfo(string path);

    /// @}

  private:
    //
    PA  offset_; ///< physical address offset
    int size_;   ///< physical memory size
};

//------------------------------------------------------------------------------
/// @brief parallel application
///
/// parallel application description
///
class ParApp {
  public:
    // scheduler can access any methods of application
    friend class CoreAllocator;

    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    ///
    /// @param pid          application id
    /// @param name         application name
    /// @param scheduler    pointer to scheduler
    /// @param architecture pointer to architecture instance
    ParApp(pid_t pid, string name,
        CoreAllocator* scheduler, CPUArchitecture* architecture);

    /// @brief destructor
    ~ParApp();

    /// @}

    /// @name application attributes
    /// @{

    void SetDelegate(int id) { delegate_id_ = id; }
    int GetDelegate(void) { return delegate_id_; }

    pid_t pid(void) const { return pid_; }
    string name(void) const { return name_; }

    /// @brief get the allocated tile number
    int GetNumTiles(void) { return allocated_tiles_; }

    /// @brief get the list of tiles allocated for the application
    list<CPUTile*> GetTileSet(void) const { return tile_list_; }
    list<CPUTile*> GetAllocatedTileSet(void) const
    { return allocated_tile_list_; }

    /// @brief get the cpuset (cpu bitmap) from application tile list
    rte_cpu_set GetRTECPUSet(void);

    /// @brief update application parallel context through shared memory area
    void UpdateAppContext(void);

    /// @brief get current application parallel region
    int GetParallelRegion(void) { return current_region_; }

    /// @brief caching current region to cache region
    void CachingCurrentRegion(void) { cached_region_ = current_region_; }

    /// @brief get caching region number
    int GetCachedRegion(void) { return cached_region_; }

    /// @brief returns the current CPU utilization by ParApp instance
    double GetUtil();

    /// @brief update LLC miss rate per tile of given parallel region
    void UpdateLLCAccessRate(unsigned long llc_accesses, unsigned long total_cycles);
    void UpdateLLCMissRate(unsigned long llc_misses, unsigned long total_cycles);
    void UpdateLLCMissAccessRate(unsigned long llc_misses, unsigned long llc_accesses);

    double GetLLCMissAccessRate(void);

    void SetRegionProfiled(void);
    bool IsRegionProfiled(void);

    /// @brief update DRAM request rate per tile of given parallel region
    ///
    /// @param DRAM requests
    /// @param total cycles
    void UpdateDRAMReqRate(unsigned long dram_reqs, unsigned long total_cycles);

    /// @brief update total cycles of a given parallel region
    ///
    /// @param total cycles
    void UpdateCurrCpuCycles(uint64_t total_cycles);

    /// @brief dump applicatio performance profiles
    void DumpPerformanceProfiles(string path);

    /// @brief show application resources info.
    void ShowAppResource(ostream & dout);

    /// @brief write detailed application info. into a log file
    void ShowAppInfo(string path);

    /// @brief inform application context (cpu set) to appication runtime
    void ScheduleCPUs(void);

    /// @brief allocate one tile
    void AllocOneTile(CPUTile* tile);

    /// @brief deallocate one tile
    bool DeAllocOneTile(CPUTile* tile);

    bool GetNeedProfiling(void) { return need_profiling_; }
    void SetProfiling(void) { need_profiling_ = true; }
    void UnsetProfiling(void) { need_profiling_ = false; }

    void SetProfilingNode(CPUNode* profiling_node) {
        profiling_node_ = profiling_node;
    }

    void UnsetProfilingNode(void) {
        profiling_node_ = NULL;
    }

    CPUNode* GetProfilingNode(void) {
        return profiling_node_;
    }

    /// @}

  private:
    /// @name application private attributes
    /// @{

    /// @brief flush all allocated tiles
    bool FlushAllTiles(void);

    /// @brief get llc miss rate of given number of tiles
    double GetLLCMissRate(int tiles);

    /// @brief get dram request rate of given number of tiles
    double GetDRAMReqRate(int tiles);

    /// @brief get total cycles
    unsigned long GetCurrCpuCycles();

    /// @brief initialize shared memory area
    /// @description it is used to communicate application context
    void InitSHM(void);

    /// @brief cleanup the shared memory area
    void ReleaseSHM(void);

    /// @}

    // application info.
    const pid_t  pid_;  ///< application id
    const string name_; ///< application name

    int delegate_id_;

    // core resources
    list<CPUTile*> tile_list_;  ///< the allocated tile instances
    list<CPUTile*> allocated_tile_list_;
    int allocated_tiles_;       ///< the number of allocated tiles
    int reserved_tiles_;
    int reserved_nodes_;

    CPUNode* profiling_node_;

    uint64_t curr_cpu_cycles_;

    // performance model
    map<int, RegionProfiles*> profiles_;
    char current_kernel_name_[4096];
    int current_region_; ///< current parallel region number
    int cached_region_;

    // shared memory area between application runtime and RTE
    int   shmid_;      ///< shared memory id
    key_t key_;        ///< shared memory key
    void* shmaddr_;    ///< shared memory address
    bool  shm_opened_; ///< shared memory is initialized
    int* alloc_map_; ///< core allocation map for application runtime
    vector<int> numa_node_ids;
    int** node_alloc_map_; ///< per-node core alloc. map for application runtime

    CoreAllocator*    scheduler_; ///< the scheduler instance
    CPUArchitecture*  arch_;      ///< the target architecture instance

    bool need_profiling_;
};

#endif
