//------------------------------------------------------------------------------
/// @brief RTE - application runtime communication protocol implementation
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription app_comm.cpp
/// communication protocol implementation.
/// implementation of how an application runtime and RTE interact each other.
/// 
//------------------------------------------------------------------------------

#include "app_comm.h"

//------------------------------------------------------------------------------
/// @brief Application Runtime Communication Socket Version Protocol Definition
RTEAppConnection::RTEAppConnection(int port)
    : port_(port)
{
    rte_socket_opened_ = false;
    app_socket_opened_ = false;
}

RTEAppConnection::~RTEAppConnection()
{
  if (rte_socket_opened_) {
    close(rte_socket_);
  }

  if (app_socket_opened_) {
    close(app_socket_);
  }
}

int RTEAppConnection::Socket(void)
{
  int optval = 1;

  if ((rte_socket_ = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("socket error");
    return -1;
  }
  
  if (setsockopt(rte_socket_, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) < 0) {
    perror("setsockopt error");
    return -1;
  }

  memset((void*)&sockaddr, 0x00, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  sockaddr.sin_port = htons(port_);

  rte_socket_opened_ = true;

  return 0;
}

int RTEAppConnection::Bind(void)
{
  // quick fix: qualify bind to global namespace
  // TODO: fix using namespace std; among overall codes
  if (::bind(rte_socket_, (struct sockaddr*)&sockaddr,
        sizeof(sockaddr)) == -1) {
    perror("bind error");
    return -1;
  }

  return 0;
}

int RTEAppConnection::Listen(void)
{
  if (listen(rte_socket_, 1) == -1) {
    perror("listen error");
    return -1;
  }

  return 0;
}

int RTEAppConnection::Accept(void)
{
  socklen_t socklen;

  if (app_socket_opened_) {
    perror("unexpected situation: application socket already opened");
    return -1;
  }

  app_socket_ = accept(rte_socket_, (struct sockaddr*)&sockaddr, &socklen);
  if (rte_socket_ == -1) {
    perror("accept error");
    return -1;
  }

  app_socket_opened_ = true;

  return 0;
}

int RTEAppConnection::ReadCMD(void)
{
  if (app_socket_opened_) {
    int command = -1;

    if (read(app_socket_, &command, sizeof(int)) <= 0) {
      perror("read error (command number error)");
      return -1;
    }

    /* check the command is valid */
    if (command >= APP_BEGIN && command <= APP_END)
      return command;
    else
      return -1;
  }

  return -1;
}

void RTEAppConnection::ReadMSG(int command, void* arg)
{
  switch (command) {
    case APP_BEGIN:
    {
      app_begin_data data;

      if (app_socket_opened_) {
        if (read(app_socket_, &data, sizeof(app_begin_data)) <= 0) {
          perror("read app msg data error");
          exit(-1);
        }

        memcpy(arg, &data, sizeof(app_begin_data));
      }
      break;
    }
    case APP_END:
    {
      app_end_data data;

      if (app_socket_opened_) {
        if (read(app_socket_, &data, sizeof(app_end_data)) <= 0) {
          perror("read app msg data error");
          exit(-1);
        }

        memcpy(arg, &data, sizeof(app_end_data));
      }
      break;
    }
  }
}

void RTEAppConnection::SendACK(void)
{
  if (app_socket_opened_) {
    int ack = 1;

    if (write(app_socket_, &ack, sizeof(int)) <= 0) {
      perror("ack write error");
      exit(-1);
    }
  }
}

void RTEAppConnection::Close(void)
{
  if (app_socket_opened_) {
    close(app_socket_);
    app_socket_opened_ = false;
  }
}

