//------------------------------------------------------------------------------
/// @brief RTE - application runtime communication protocol header
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription app_comm.h
/// communication protocol header description.
/// it describes how an application runtime and RTE interact each other.
/// 
//------------------------------------------------------------------------------

#ifndef __APP_COMM_H_
#define __APP_COMM_H_

#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <iostream>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <mutex>
#include "app.h"
#include "comm_interface.h"
#include "rte_cpu.h"
#include "common.h"

#define START_APP 1
#define ENTER_PAR_REGION 2
#define FINISH_PAR_REGION 3
#define FINISH_APP 4

using std::cout;
using std::endl;
using std::mutex;

//------------------------------------------------------------------------------
/// @brief Application Runtime Communication Socket Version Protocol
/// 
class RTEAppConnection {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    ///
    /// @param port socket port number
    RTEAppConnection(int port);

    /// @brief destructor
    ~RTEAppConnection();

    /// @}

    /// @name communication (app->RTE) protocol
    /// @{

    /// @brief socket initialization
    int Socket(void);

    /// @brief socket bind
    int Bind(void);

    /// @brief socket listen
    int Listen(void);

    /// @brief socket accept
    int Accept(void);

    /// @brief read command from application runtime
    virtual int ReadCMD(void);

    /// @brief read message from application
    ///
    /// @param command applicatioin command
    /// @param arg     message packing structure
    /// @description   save application's message into the parameter
    virtual void ReadMSG(int command, void* arg);

    /// @brief send ACK message to application
    void SendACK(void);

    /// @brief close application socket
    void Close(void);

    /// @}

  private:
    const int port_;                ///< port number
    int    rte_socket_;          ///< rte socket
    int    app_socket_;          ///< application runtime socket connection

    struct sockaddr_in sockaddr; ///< socket properties
    bool   rte_socket_opened_;   ///< rte socket status
    bool   app_socket_opened_;   ///< application socket connection status
};

#endif // __APP_COMM_H_

