//------------------------------------------------------------------------------
/// @brief ARM-RTE architecture description
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription arch.cpp
/// The implementation of target architecture instance.
///
//------------------------------------------------------------------------------

#include "arch.h"
#include "app.h"

using namespace std;

//------------------------------------------------------------------------------
CPUCore::CPUCore(int id, char* vendor, int max_clock_frequency_MHz)
    : id_(id), vendor_(vendor), max_clock_frequency_MHz_(max_clock_frequency_MHz)
{
}

CPUCore::~CPUCore()
{
}

//------------------------------------------------------------------------------
CPUTile::CPUTile(int id)
    : id_(id)
{
    colored_ = false;
    app_ = 0;

    total_cycles_epoch_ = 0;
    stall_cycles_epoch_ = 0;
}

CPUTile::~CPUTile()
{
}

bool CPUTile::RegisterApp(pid_t pid)
{
    if (IsColored()) {
        cout << "already allocated... pid(" << GetAppID() << ")" << endl;
        return false;
    }

    app_ = pid;
    colored_ = true;

    return true;
}

void CPUTile::CleanApp(void)
{
    app_ = 0;
    colored_ = false;
}

//------------------------------------------------------------------------------
CPUNode::CPUNode(int id, int num_memory_nodes)
    : id_(id), num_memory_nodes_(num_memory_nodes)
{
    distance_ = (int*)malloc(num_memory_nodes_ * sizeof(int));

    llc_accesses_epoch_ = 0;
    llc_misses_epoch_ = 0;
    dram_requests_epoch_ = 0;

    llc_access_rate_epoch_ = 0;
    llc_miss_rate_epoch_ = 0;
    dram_req_rate_epoch_ = 0;
}

CPUNode::~CPUNode()
{
    free(distance_);
}

//------------------------------------------------------------------------------
MemoryPage::MemoryPage()
  : size_(1048576), buddy_(NULL), app_(0)
{
}

MemoryPage::MemoryPage(int size, MemoryPage* buddy)
  : size_(size), buddy_(buddy), app_(0)
{
}

MemoryPage::~MemoryPage()
{
}

//------------------------------------------------------------------------------
MemoryNode::MemoryNode(int id, int size, char* type)
    : id_(id), size_(size), type_(type)
{
    num_pages_ = size_ * (1024/4);
    page_ = (MemoryPage**)malloc(num_pages_ * sizeof(MemoryPage*));

    // construct each page instance
    for (int pfn = 0; pfn < num_pages_; pfn++) {
        page_[pfn] = new MemoryPage();
    }
}

MemoryNode::~MemoryNode()
{
    for (int pfn = 0; pfn < num_pages_; pfn++)
        delete page_[pfn];

    delete page_;
}

void MemoryNode::InitPage(PA pfn, pid_t pid)
{
    if (pfn >= 0 || pfn <= num_pages_) {
        page_[pfn]->RegisterAppID(pid);
    }
}

void MemoryNode::Dump(ostream & dout)
{
    dout << "----------------------------------------"
        << "----------------------------------------" << endl;
    dout << "memory id: " << id() << endl;
    dout << "memory size: " << size() << "MB" << endl;
    dout << "memory type: " << type() << endl;

    for (int pfn = 0; pfn < num_pages_; pfn++) {
        dout << "pfn: " << pfn << " running app: "
            << page_[pfn]->GetAppID() << endl;
    }

    return;
}

void MemoryNode::MemoryInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "----------------------------------------"
        << "----------------------------------------" << endl;
    out << "detailed memory info. " << endl;
    out << "memory id: " << id() << endl;
    out << "memory size: " << size() << "MB" << endl;
    out << "memory type: " << type() << endl;

    out.close();

    return;
}

//------------------------------------------------------------------------------
static size_t PipeRead(const char* filename, char* buf, size_t size)
{
    FILE* fp = popen(filename, "r");
    size_t read_size = fread(buf, sizeof(char), size, fp);
    pclose(fp);
    return read_size;
}

static char* get_vendor_name()
{
    char buf[64];
    PipeRead("/bin/cat /proc/cpuinfo | grep '^vendor_id' | head -1", buf,
             sizeof(buf) - 1);
    char* vendor = (char*)malloc(sizeof(char) * 64);
    sscanf(buf, "vendor_id %*s %s", vendor);
    return vendor;
}

static int get_max_clock_frequency()
{
    char buf[8] = {0,};
    PipeRead("/bin/cat /sys/devices/system/cpu/cpu0/cpufreq/scaling_available_frequencies",
            buf, sizeof(buf) - 1);
    int frequency;
    sscanf(buf, "%d %*s", &frequency);
    return (frequency / 1000);
}

static int get_total_system_cores()
{
    char buf[8];
    PipeRead("/bin/cat /sys/devices/system/cpu/present", buf, sizeof(buf) - 1);
    int from, to;
    sscanf(buf, "%d-%d", &from, &to);
    int num_cores = to - from + 1;
    std::cout << num_cores << " cores are detected" << std::endl;

    return num_cores;
}

static int get_total_memory_size_kb()
{
    char buf[24];
    PipeRead("/bin/cat /proc/meminfo | grep '^MemTotal:'", buf,
            sizeof(buf) - 1);
    int memsize;
    sscanf(buf, "MemTotal %*s %d", &memsize);

    return memsize;
}

static int get_total_memory_size_mb()
{
    int total_memory_size_kb = get_total_memory_size_kb();
    int total_memory_size_mb = total_memory_size_kb / 1024;

    return total_memory_size_mb;
}

static int get_node_memory_size_mb(int node)
{
    char command[128];
    char buf[24];

    sprintf(command, "%s %d %s",
            "/usr/bin/numactl --hardware | grep 'node",
            node,
            "size:'");

    PipeRead(command, buf, sizeof(buf) - 1);

    int memsize;
    sscanf(buf, "node %*s %*s %d %*s", &memsize);

    return memsize;
}

static int get_distance_from_to(int cpu, int memory)
{
    char command[128] = { 0, };
    char buf[64] = { 0, };

    sprintf(command, "%s%d:'", "/usr/bin/numactl --hardware | grep '", cpu);

    PipeRead(command, buf, sizeof(buf) - 1);

    char query_str[128] = { 0, };
    char prefix[12] = { 0, };
    sprintf(prefix, "  %d:  ", cpu);
    strncat(query_str, prefix, strlen(prefix));
    int i = 0;
    while (i < memory) {
        strncat(query_str, "%*s  ", 5);
        i++;
    }
    strncat(query_str, "%d  ", 4);

    int distance;
    sscanf(buf, query_str, &distance);

    return distance;
}

CPUArchitecture::CPUArchitecture()
{
    sprintf(vendor_, "%s", get_vendor_name());
    sprintf(memory_type_, "%s", "dram");
    max_clock_frequency_MHz_ = get_max_clock_frequency();
    total_memory_size_MB_ = get_total_memory_size_mb();

    if (numa_available() == -1) {
        sprintf(arch_type_, "%s", "UMA");
        BuildUMAArchitecture();
    } else {
        sprintf(arch_type_, "%s", "NUMA");
        BuildNUMAArchitecture();
    }

    alloc_map_ = (int*)calloc(num_tiles_, sizeof(int));
    alloc_cached_ = (int*)calloc(num_tiles_, sizeof(int));
}

CPUArchitecture::~CPUArchitecture()
{
    for (int i = 0; i < num_cores_; i++)
        free(cpu_cores_[i]);
    free(cpu_cores_);
    for (int i = 0; i < num_tiles_; i++)
        free(cpu_tiles_[i]);
    free(cpu_tiles_);
    for (int i = 0; i < num_cpu_nodes_; i++)
        free(cpu_nodes_[i]);
    free(cpu_nodes_);

    free(alloc_map_);
    free(alloc_cached_);
}

CPUNode* CPUArchitecture::GetClosestCPUNode(int memory_node_id)
{
    int distance = 100;
    int node = 0;

    for (int i = 0; i < num_cpu_nodes_; i++) {
        if (cpu_nodes_[i]->distance(memory_node_id) < distance) {
            distance = cpu_nodes_[i]->distance(memory_node_id);
            node = i;
        }
    }

    return cpu_nodes_[node];
}

bool CPUArchitecture::LookupAllocationCache()
{
    /* get new scheduling result */
    for (int i = 0; i < num_tiles_; i++) {
        alloc_map_[i] = cpu_tiles_[i]->GetAppID();
    }

    /* cache hit */
    if (!memcmp(alloc_map_, alloc_cached_, num_tiles_ * sizeof(int))) {
        return true;
    }

    ///* cache miss */
    //memcpy(alloc_cached_, alloc_map_, num_tiles_ * sizeof(int));

    return false;
}

void CPUArchitecture::UpdateAllocationCache()
{
    /* fill up */
    memcpy(alloc_cached_, alloc_map_, num_tiles_ * sizeof(int));
}

void CPUArchitecture::BuildUMAArchitecture()
{
    num_cpu_nodes_ = 1;
    num_memory_nodes_ = 1;
    num_tiles_ = get_total_system_cores();
    num_cores_ = get_total_system_cores();

    cpu_nodes_ = (CPUNode**)malloc(num_cpu_nodes_ * sizeof(CPUNode*));
    cpu_tiles_ = (CPUTile**)malloc(num_tiles_ * sizeof(CPUTile*));
    cpu_cores_ = (CPUCore**)malloc(num_cores_ * sizeof(CPUCore*));

    cpu_nodes_[0] = new CPUNode(0, num_tiles_);
    cpu_nodes_[0]->set_num_tiles(num_tiles_);
    cpu_nodes_[0]->set_num_cores(num_cores_);
    for (int i = 0; i < num_tiles_; i++) {
        cpu_tiles_[i] = new CPUTile(i);
        cpu_cores_[i] = new CPUCore(i, vendor_, max_clock_frequency_MHz_);

        cpu_tiles_[i]->SetCPUNode(cpu_nodes_[0]);
        cpu_tiles_[i]->SetNodeIndex(i);

        cpu_cores_[i]->SetCPUNode(cpu_nodes_[0]);
        cpu_cores_[i]->SetCPUTile(cpu_tiles_[i]);
        cpu_cores_[i]->SetNodeIndex(i);
        cpu_cores_[i]->SetTileIndex(0);

        cpu_nodes_[0]->SetTile(cpu_tiles_[i]);
        cpu_tiles_[i]->SetCore(cpu_cores_[i]);
    }

    memory_nodes_ =
        (MemoryNode**)malloc(num_memory_nodes_ * sizeof(MemoryNode*));
    memory_nodes_[0] = new MemoryNode(0, total_memory_size_MB_, memory_type_);

    cpu_nodes_[0]->SetDistance(0, get_distance_from_to(0, 0));
}

void CPUArchitecture::BuildNUMAArchitecture()
{
    num_cpu_nodes_ = numa_num_task_nodes();
    num_memory_nodes_ = numa_num_task_nodes(); // TODO
    num_tiles_ = numa_num_task_cpus();
    num_cores_ = get_total_system_cores();

    cpu_nodes_ = (CPUNode**)malloc(num_cpu_nodes_ * sizeof(CPUNode*));
    cpu_tiles_ = (CPUTile**)malloc(num_tiles_ * sizeof(CPUTile*));
    cpu_cores_ = (CPUCore**)malloc(num_cores_ * sizeof(CPUCore*));

    std::vector<std::vector<int> > available_cus_;

    int numa_node_checked = 0;
    int node_id = 0;

    while (numa_node_checked < num_cpu_nodes_) {
        struct bitmask* available_cpus_bitmap;
        available_cpus_bitmap = numa_allocate_cpumask();
        if (numa_node_to_cpus(node_id, available_cpus_bitmap)) {
            /* system numa node id is not necessarily linear */
            node_id++;
            continue;
        }

        node_ids_.push_back(node_id);
        node_id++;
        numa_node_checked++;

        std::vector<int> available_cpus;
        for (int cpu = 0; cpu < numa_num_task_cpus(); cpu++)
            if (numa_bitmask_isbitset(available_cpus_bitmap, cpu))
                available_cpus.push_back(cpu);

        available_cus_.push_back(available_cpus);

        numa_free_cpumask(available_cpus_bitmap);
    }

    int cpu_index = 0;
    for (int i = 0; i < num_cpu_nodes_; i++) {
        cpu_nodes_[i] = new CPUNode(node_ids_[i], num_memory_nodes_);
        cpu_nodes_[i]->set_num_tiles(available_cus_[i].size());
        cpu_nodes_[i]->set_num_cores(available_cus_[i].size());

        for (int j = 0; j < available_cus_[i].size(); j++) {
            int cpu = available_cus_[i][j];
            cpu_tiles_[cpu_index] = new CPUTile(cpu);
            cpu_cores_[cpu_index] = new CPUCore(cpu,
                    vendor_, max_clock_frequency_MHz_);

            cpu_tiles_[cpu_index]->SetCPUNode(cpu_nodes_[i]);
            cpu_tiles_[cpu_index]->SetNodeIndex(j);

            cpu_cores_[cpu_index]->SetCPUNode(cpu_nodes_[i]);
            cpu_cores_[cpu_index]->SetCPUTile(cpu_tiles_[cpu_index]);
            cpu_cores_[cpu_index]->SetNodeIndex(j);
            cpu_cores_[cpu_index]->SetTileIndex(0);

            cpu_nodes_[i]->SetTile(cpu_tiles_[cpu_index]);
            cpu_tiles_[cpu_index]->SetCore(cpu_cores_[cpu_index]);
            cpu_index++;
        }
    }

    memory_nodes_ =
        (MemoryNode**)malloc(num_memory_nodes_ * sizeof(MemoryNode*));
    for (int i = 0; i < num_memory_nodes_; i++) {
        memory_nodes_[i] = new MemoryNode(i,
                get_node_memory_size_mb(i),
                memory_type_);
    }

    for (int i = 0; i < num_cpu_nodes_; i++) {
        for (int j = 0; j < num_memory_nodes_; j++) {
            cpu_nodes_[i]->SetDistance(j, get_distance_from_to(i, j));
        }
    }
}

void CPUArchitecture::BriefArchitectureInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "Brief Architecture Information" << endl;
    out << "arch type: " << arch_type_ << endl;
    out << "arch vendor: " << vendor_ << endl;
    out << "# CPU nodes: " << num_cpu_nodes_ << endl;
    out << "# CPU tiles: " << num_tiles_ << endl;
    out << "# CPU cores: " << num_cores_ << endl;
    out << "# memory nodes: " << num_memory_nodes_ << endl;
    out << "total memory size: " << total_memory_size_MB_ << "MB" << endl;

    out.close();

    return;
}

void CPUArchitecture::DetailedArchitectureInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "Detailed Architecture Information" << endl;
    out << "arch type: " << arch_type_ << endl;
    out << "arch vendor: " << vendor_ << endl;
    out << "# CPU nodes: " << num_cpu_nodes_ << endl;
    out << "# CPU tiles: " << num_tiles_ << endl;
    out << "# CPU cores: " << num_cores_ << endl;
    out << "# memory nodes: " << num_memory_nodes_ << endl;
    out << "total memory size: " << total_memory_size_MB_ << "MB" << endl;

    out << "========================================"
        << "========================================" << endl;
    out << "Memory Information" << endl;

    for (int i=0; i<num_memory_nodes(); i++)
        GetMemoryNode(i)->MemoryInfo(path);

    // detailed information
    for (int i=0; i<num_cpu_nodes_; i++) {
        CPUNode* cpu_node = cpu_nodes_[i];
        out << "----------------------------------------"
            << "----------------------------------------" << endl;
        out << "CPU node information" << endl;
        out << "node " << i << " id: " << cpu_node->id() << endl;
        out << "the number of tiles that this node contains: "
            << cpu_node->num_tiles() << endl;

        out << "the distance to the memory node" << endl;
        for (int j=0; j<num_memory_nodes_; j++) {
            out << "    node: " << j
                << " distance: " << cpu_node->distance(j) << endl;
        }

        for (int j=0; j<cpu_node->num_tiles(); j++) {
            CPUTile* cpu_tile = cpu_node->GetTile(j);
            out << "----------------------------------------"
                << "----------------------------------------" << endl;
            out << "  Tile information (in cluster " << i << ")" << endl;
            out << "  tile " << j << " id: " << cpu_tile->id() << endl;

            out << "  the number of cores that this tile contains: "
                << cpu_tile->num_cores() << endl;

            for (int k=0; k<cpu_tile->num_cores(); k++) {
                CPUCore* cpu_core = cpu_tile->GetCore(k);
                out << "    Core information (in tile " << j << ")" << endl;
                out << "    core " << k << " id: " << cpu_core->id()
                    << " vendor type: " << cpu_core->vendor()
                    << " max clock frequency: "
                    << cpu_core->max_clock_frequency_MHz()
                    << "MHz"
                    << endl;
            }

            out << endl;
        }
    }

    out.close();

    return;
}


