//------------------------------------------------------------------------------
/// @brief ARM-RTE architecture description
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription arch.h
/// The description of target architecture instance.
/// We model CPUs with three different CPU types in a hierachical manner.
///
/// CPU node: a set of Tiles. A CPU node shares an LLC (last-level cache)
/// CPU tile: a set of Cores. The cores belonging to a same tile have same
///           memory access latency. Default core allocation unit of RTE.
/// CPU core: a basic compute unit.
///
//------------------------------------------------------------------------------

#ifndef __ARCH_DESCRIPTION_
#define __ARCH_DESCRIPTION_

#include <cassert>
#include <cstdio>
#include <cstring>
#include <fstream>
#include <iostream>
#include <vector>
#include <map>

#include <numa.h>
#include <numaif.h>

#define CPA     int
#define PA      int

#define MEMOFFSET  100
#define MEMMASK(x) x/MEMOFFSET

class CPUNode;
class CPUTile;

//------------------------------------------------------------------------------
/// @brief Core description
///
/// basic description of core structure
/// a core is a basic computing unit
///
class CPUCore {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    CPUCore(int id, char* vendor, int max_clock_frequency_MHz);

    /// @brief destructor
    ~CPUCore();

    /// @}

    /// @name core attributes
    /// @{

    int id(void) const { return id_; }
    char* vendor(void) const { return vendor_; }
    int max_clock_frequency_MHz(void) const { return max_clock_frequency_MHz_; }

    void SetCPUNode(CPUNode* cpu_node) { cpu_node_ = cpu_node; }
    CPUNode* GetCPUNode(void) { return cpu_node_; }

    void SetCPUTile(CPUTile* cpu_tile) { cpu_tile_ = cpu_tile; }
    CPUTile* GetCPUTile(void) { return cpu_tile_; }

    void SetNodeIndex(int idx) { node_idx_ = idx; }
    int GetNodeIndex(void) { return node_idx_; }

    void SetTileIndex(int idx) { tile_idx_ = idx; }
    int GetTileIndex(void) { return tile_idx_; }

    /// @}

  private:
    CPUNode* cpu_node_; ///< parent CPU node
    CPUTile* cpu_tile_; ///< parent CPU Tile

    int node_idx_; ///< index in CPU node
    int tile_idx_; ///< index in CPU tile

    const int id_; ///< core id
    char* vendor_; ///< vendor name
    const int max_clock_frequency_MHz_; ///< core maximum clock frequency (in MHz)
};

//------------------------------------------------------------------------------
/// @brief Tile description
///
/// basic description of tile structure
/// a tile is a set of cores
/// the cores in a same tile have same memory access latency
/// a tile is a basic granularity of scheduling (i.e., core allocation)
///
class CPUTile {
  public:
    // architecture instance can access any methods of tile
    friend class CPUArchitecture;

    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    CPUTile(int id);

    /// @brief destructor
    ~CPUTile();

    /// @}

    /// @name tile attributes
    /// @{

    int id(void) const { return id_; }
    int num_cores(void) { return cores_.size(); }

    void SetCPUNode(CPUNode* cpu_node) { cpu_node_ = cpu_node; }
    CPUNode* GetCPUNode(void) { return cpu_node_; }

    void SetNodeIndex(int idx) { node_idx_ = idx; }
    int GetNodeIndex(void) { return node_idx_; }

    /// @brief get core instance of given id
    CPUCore* GetCore(int index) { return cores_[index]; }

    /// @brief check whether the tile is allocated or not
    bool IsColored(void) { return colored_; }

    /// @brief register application id that runs on this tile
    bool RegisterApp(pid_t pid);

    /// @brief remove any application id on the tile
    void CleanApp(void);

    /// @brief get application id that runs on this tile
    pid_t GetAppID(void) { return app_; }

    void UpdateTotalCycles_Epoch(unsigned long total_cycles) {
        total_cycles_epoch_ = total_cycles;
    }

    void UpdateStallCycles_Epoch(unsigned long stall_cycles) {
        stall_cycles_epoch_ = stall_cycles;
    }

    void UpdateCPUUtilization(double cpu_utilization) {
        cpu_utilization_ = cpu_utilization;
    }

    void UpdateLLCMissRate(double llc_miss_rate) {
        llc_miss_rate_ = llc_miss_rate;
    }

    unsigned long GetTotalCycles_Epoch(void) {
        return total_cycles_epoch_;
    }

    unsigned long GetStallCycles_Epoch(void) {
        return stall_cycles_epoch_;
    }

    double GetCPUUtilization(void) {
        return cpu_utilization_;
    }

    double GetLLCMissRate() { return llc_miss_rate_; }

    int Lookup(void) {
        if (lookup_count_ == 5) {
            return 1;
        } else {
            lookup_count_++;
            return 0;
        }
    }

    void LookupClean() { lookup_count_ = 0; }

    int SetLLCExchangeFlag() { llc_exchange_possible = 1; }
    void CleanLLCExchangeFlag() { llc_exchange_possible = 0; }
    int GetLLCExchangeFlag() { return llc_exchange_possible; }

    /// @}

  private:
    /// @name tile private attributes
    ///
    /// @{

    /// @brief link between tile and cores
    void SetCore(CPUCore* core) { cores_.push_back(core); }
    void SetCore(int index, CPUCore* core) { cores_[index] = core; }

    /// @}

    const int id_;  ///< tile id within the node

    CPUNode* cpu_node_; ///< parent CPU node
    int node_idx_; ///< index in CPU node

    std::vector<CPUCore*> cores_; ///< core instances

    bool      colored_;   ///< flag to indicate idle tile
    pid_t     app_;       ///< the application id that runs on this tile

    unsigned long total_cycles_epoch_ = 0;
    unsigned long stall_cycles_epoch_ = 0;
    double cpu_utilization_ = 0;
    double llc_miss_rate_ = 0;

    int lookup_count_ = 0;

    int llc_exchange_possible = 0;
};

//------------------------------------------------------------------------------
/// @brief node description
///
/// basic description of node structure
/// a node is a set of tiles
///
class CPUNode {
  public:
    // architecture instance can access any methods of node
    friend class CPUArchitecture;

    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    CPUNode(int id, int num_memory_nodes);

    /// @brief destructor
    ~CPUNode();

    /// @}

    /// @name node attributes
    /// @{

    int id(void) { return id_; }

    void set_num_tiles(int num_tiles) { num_tiles_ = num_tiles; }
    void set_num_cores(int num_cores) { num_cores_ = num_cores; }
    int num_tiles(void) { return num_tiles_; }
    int num_cores(void) { return num_cores_; }

    /// @brief get tile instance of given id
    CPUTile* GetTile(int index) { return tiles_[index]; }

    /// @brief get the maximum distance that system has
    int max_distance(void) { return max_distance_; }

    /// @brief get the distance to the memory node
    int distance(int memory_node_id) { return distance_[memory_node_id]; }

    /// @brief update performance measurement data during the latest epoch

    void UpdateTotalCycles_Epoch(unsigned long total_cycles) {
        total_cycles_epoch_ = total_cycles;
    }

    void UpdateStallCycles_Epoch(unsigned long stall_cycles) {
        stall_cycles_epoch_ = stall_cycles;
    }

    void UpdateLLCAccesses_Epoch(unsigned long llc_accesses) {
        llc_accesses_epoch_ = llc_accesses;
    }

    void UpdateLLCMisses_Epoch(unsigned long llc_misses) {
        llc_misses_epoch_ = llc_misses;
    }

    void UpdateDRAMRequests_Epoch(unsigned long dram_requests) {
        dram_requests_epoch_ = dram_requests;
    }

    unsigned long GetTotalCycles_Epoch(void) {
        return total_cycles_epoch_;
    }

    unsigned long GetStallCycles_Epoch(void) {
        return stall_cycles_epoch_;
    }

    unsigned long GetLLCAccesses_Epoch(void) {
        return llc_accesses_epoch_;
    }

    unsigned long GetLLCMisses_Epoch(void) {
        return llc_misses_epoch_;
    }

    unsigned long GetDRAMRequests_Epoch(void) {
        return dram_requests_epoch_;
    }

    void UpdateLLCAccessRate_Epoch(unsigned long llc_accesses, unsigned long total_cycles) {
        llc_access_rate_epoch_ = (double)llc_accesses / (double)total_cycles;
    }

    void UpdateLLCMissRate_Epoch(unsigned long llc_misses, unsigned long total_cycles) {
        llc_miss_rate_epoch_ = (double)llc_misses / (double)total_cycles;
    }

    void UpdateDRAMReqRate_Epoch(unsigned long dram_requests, unsigned long total_cycles) {
        dram_req_rate_epoch_ = (double)dram_requests / (double)total_cycles;
    }

    double GetLLCAccessRate_Epoch(void) {
        return llc_access_rate_epoch_;
    }

    double GetLLCMissRate_Epoch(void) {
        return llc_miss_rate_epoch_;
    }

    double GetDRAMReqRate_Epoch(void) {
        return dram_req_rate_epoch_;
    }

    /// @}

  private:
    /// @name node private attributes
    ///
    /// @{

    /// @brief link between node and tiles
    void SetTile(CPUTile* tile) { tiles_.push_back(tile); }

    /// @breif set distance between CPU node and memory node
    void SetDistance(int memory_node_id, int distance) {
        assert(memory_node_id >= 0 && memory_node_id < num_memory_nodes_);
        distance_[memory_node_id] = distance;
    }

    /// @}

    const int id_;          ///< node id within the system

    int* distance_;
    int  num_memory_nodes_; ///< the number of memory nodes
    int  max_distance_;

    int num_tiles_;
    int num_cores_;

    std::vector<CPUTile*> tiles_;

    unsigned long total_cycles_epoch_;
    unsigned long stall_cycles_epoch_;

    unsigned long llc_accesses_epoch_;
    unsigned long llc_misses_epoch_;
    unsigned long dram_requests_epoch_;

    double llc_access_rate_epoch_;
    double llc_miss_rate_epoch_;
    double dram_req_rate_epoch_;
};

//------------------------------------------------------------------------------
/// @brief MemoryPage description
///
/// basic description of page structure
/// page is basic system(physical) memory management granularity
/// in our prototype, each page is 1MB of memory
/// (e.g., 4096 pages -> 4GB of memory)
///
class MemoryPage {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief default constructor
    MemoryPage();

    /// @brief constructor taking initialization values
    ///
    /// @param size  page size
    /// @param buddy pointer to a page that is adjacent with this page
    MemoryPage(int size, MemoryPage* buddy);

    /// @brief destructor
    ~MemoryPage();

    /// @}

    /// @name page attributes
    /// @{

    /// @brief register application
    void RegisterAppID(pid_t pid) { app_ = pid; }

    /// @brief get application id
    pid_t GetAppID(void) { return app_; }

    /// @}

  private:
    //
    const int    size_;  ///< page size: default size is 1048576 bytes (1MB)
    const MemoryPage* buddy_; ///< one future consideration: linux buddy style mgmt.
    pid_t        app_;   ///< application id
};

//------------------------------------------------------------------------------
/// @brief MemoryNode description
///
/// basic description of memory structure
///
class MemoryNode {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    ///
    /// @param id   memory node id
    /// @param size the size of memory
    /// @param type the type of memory (e.g., ddr3)
    MemoryNode(int id, int size, char* type);

    /// @brief destructor
    ~MemoryNode();

    /// @}

    /// @name memory attributes
    /// @{

    int id(void) const { return id_; }
    int size(void) const { return size_; }
    char* type(void) const { return type_; }

    /// @brief get page from given page number
    ///
    /// @param  pfn page frame number
    /// @retval pointer to page instance
    MemoryPage* GetPage(int pfn) { return page_[pfn]; }

    /// @brief initialize page with application id
    ///
    /// @param pfn page frame number
    /// @param pid application id
    void InitPage(PA pfn, pid_t pid);

    /// @brief simple memory dump
    void Dump(std::ostream & out);

    /// @brief write detailed memory info. into a log file
    void MemoryInfo(std::string path);

    /// @}

  private:
    const int   id_;   ///< memory node id
    const int   size_; ///< memory size (MB)
    char* type_; ///< memory type

    int num_pages_;
    MemoryPage**      page_; ///< page instances of the memory
};

//------------------------------------------------------------------------------
/// @brief CPUArchitecture description
///
/// basic description of prototype target architecture
///
class CPUArchitecture {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief default constructor
    CPUArchitecture();

    /// @brief destructor
    ~CPUArchitecture();

    /// @}

    /// @name architecture attributes
    /// @{

    char* vendor(void) { return vendor_; }
    int max_clock_frequency_MHz(void) const { return max_clock_frequency_MHz_; }
    char* arch_type(void) { return arch_type_; }
    std::vector<int> node_ids() { return node_ids_; }

    int num_cpu_nodes(void) const { return num_cpu_nodes_; }
    int num_memory_nodes(void) const { return num_memory_nodes_; }
    int num_tiles(void) const { return num_tiles_; }
    int num_cores(void) const { return num_cores_; }

    /// @brief get core instance from given id
    CPUCore* GetCore(int id) { return cpu_cores_[id]; }

    /// @brief get tile instance from given id
    CPUTile* GetTile(int id) { return cpu_tiles_[id]; }

    /// @brief get node instance from given id
    CPUNode* GetCPUNode(int id) { return cpu_nodes_[id]; }

    /// @brief get node which is closest to the given memory node
    CPUNode* GetClosestCPUNode(int memory_node_id);

    /// @brief get memory instance from given id
    MemoryNode* GetMemoryNode(int id) { return memory_nodes_[id]; }

    /// @brief check cached scheduling result and caching the new result
    ///
    /// @retval true  cache hit
    /// @retval false cache miss
    bool LookupAllocationCache(void);

    void UpdateAllocationCache(void);

    /// @brief write brief architecture info. into a log file
    void BriefArchitectureInfo(std::string path);

    /// @brief write detailed architecture info. into a log file
    void DetailedArchitectureInfo(std::string path);

    /// @}

  private:
    /// @name architecture private attributes
    /// @{

    /// @brief build architecture instance
    void BuildUMAArchitecture();
    void BuildNUMAArchitecture();

    /// @}

    //
    char vendor_[24];      ///< target core architecture vendor
    int max_clock_frequency_MHz_; ///< maximum core clock frequency
    char arch_type_[12];   ///< target architecture type
    char memory_type_[12];

    int num_cpu_nodes_;
    int num_tiles_;
    int num_cores_;
    std::vector<int> node_ids_; ///< NUMA node ids

    int num_memory_nodes_;
    int total_memory_size_MB_;

    CPUNode**  cpu_nodes_;
    CPUTile**  cpu_tiles_;
    CPUCore**  cpu_cores_;
    MemoryNode**  memory_nodes_;


    int*       alloc_map_;    ///< tile map
    int*       alloc_cached_;   ///< cached scheduled map
};

#endif

