//------------------------------------------------------------------------------
/// @brief ARM - RTE monitor
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription monitor.cpp
/// a main point of RTE framework. it acts as a daemon.
/// it communicates with application runtimes and perform scheduling through
/// scheduling module. and it manages hardware resources through shepherd
/// kernel module. application communication points are applications' each
/// fork-join point and when an application is starting or finishing.
///
//------------------------------------------------------------------------------

#include <thread>
#include <signal.h>
#include <execinfo.h>
#include "common.h"
#include "prototype.h"
#include "power_interface.h"
#include "misc.h"

//#define NUPOCO
//#define C3PO

using namespace std;

PowerInterface* power = new PowerInterface();

void signal_handler(int signal, struct sigcontext ctx)
{
    cout << "Caught signal: \'" << strsignal(signal) << "\'" << endl;

    // restore origina state
    if(likely(power))
        power->restore_orig_state();

    // print backtrace in case of segmentation fault
    if(signal == SIGSEGV)
        print_backtrace(BACKTRACE_DEPTH, ctx);

    exit(0);
}

static bool CheckAppsRegion(map<pid_t, ParApp*> apps)
{
    bool re_schedule = true;

    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;
        app->UpdateAppContext();

        std::cout << "CheckAppsRegion app: " << app->name()
            << " region: " << app->GetParallelRegion() << std::endl;

        if (app->GetCachedRegion() != app->GetParallelRegion()) {
            app->CachingCurrentRegion();
            std::cout << app->name() << " set profiling. cached_region: "
                << app->GetCachedRegion() << " current_region: "
                << app->GetParallelRegion() << std::endl;
            app->SetProfiling();
            re_schedule = false;
            //} else {
            //  app->UnsetProfiling();
    }
    }

    return re_schedule;
}

static void ScheduleAppsCPUs(map<pid_t, ParApp*> apps)
{
    std::cout << "ScheduleAppsCPUs" << std::endl;
    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;

        std::cout << app->name() << " app_region: " <<
            app->GetParallelRegion() << std::endl;
        app->ScheduleCPUs();
    }
}

void PeriodicSchedulerRoutine_C3PO(CoreAllocator* scheduler, CProfiler* profiler, CPUArchitecture* architecture)
{
    while (true) {
        /* periodically try to perform system-wide rescheduling */
        profiler->StartProfiling();

        usleep(100000); // 100ms

        if (!scheduler->RunningApps()) continue;

        profiler->StopProfiling();

        scheduler->Schedule();

        /* kernel-level system resource re-allocation */
        ScheduleAppsCPUs(scheduler->GetParApps());
        // scheduler->ShowScheduleResult("log/schedule_result.log");
    }
}

void PeriodicSchedulerRoutine(CoreAllocator* scheduler, CProfiler* profiler, CPUArchitecture* architecture)
{
    int force_schedule = 0;

    CHeuristicsProfiler* heuristics_profiler =
        dynamic_cast<CHeuristicsProfiler*>(profiler);

    while (true) {
        /* periodically try to perform system-wide rescheduling */
        if (heuristics_profiler != NULL) {
            if (heuristics_profiler->GetProfilingPhase() == CLEAN_PROFILING) {
                usleep(100000);
            }
        }
        usleep(50000); // 30 ms

        if (!scheduler->RunningApps()) continue;

        if (force_schedule > 33) { // after 1 second
            force_schedule = 0;

            cout << "force schedule" << endl;

            /*
             * try re-scheduling by taking new application performance model built
             * from offline profiler
             */
            profiler->StopProfiling();
            scheduler->Schedule();

            /*
             * check cached scheduling result. perform kernel-level resource
             * re-allocation only when cache misss happens
             */
            if (architecture->LookupAllocationCache()) {
                // start new profiling
                profiler->StartProfiling();
                continue;
            }

            /* kernel-level system resource re-allocation */
            ScheduleAppsCPUs(scheduler->GetParApps());
            //scheduler->ShowScheduleResult("log/schedule_result.log");

            /* start new profiling */
            profiler->StartProfiling();
        }

        /*
         * check whether re-scheduling is proper or not
         * : check applications' cached parallel region number and
         *   current parallel region number. try re-scheduling only when
         *   every application has same region number in both cached region number
         *   and current parallel region number
         */
        if (!CheckAppsRegion(scheduler->GetParApps())) {
            // start new profiling
            profiler->StartProfiling();
            scheduler->SetSchedulingPhase(PROFILING_RUN);
            force_schedule++;
            continue;
        }

        /*
         * try re-scheduling by taking new application performance model built
         * from offline profiler
         */
        profiler->StopProfiling();
        scheduler->Schedule();

        /* kernel-level system resource re-allocation */
        ScheduleAppsCPUs(scheduler->GetParApps());
        //scheduler->ShowScheduleResult("log/schedule_result.log");

        /* start new profiling */
        profiler->StartProfiling();

        force_schedule = 0;
    }
}

void AppConnectionListener(RTEAppConnection* app_connection, CoreAllocator* scheduler)
{
    app_connection->Socket();
    app_connection->Bind();

    cout << "Hello Dynamic RTE Monitor" << endl;

    while (true) {
        /* request from appllication runtimes */
        app_connection->Listen();
        app_connection->Accept();

        /* perform scheduling at 4 scheduling points */
        int cmd = app_connection->ReadCMD();

        switch(cmd) {
            case APP_BEGIN:
                {
                    app_begin_data msg;
                    app_connection->ReadMSG(cmd, &msg);

                    // hard coded - don't consider numactl process
                    if (!strncmp("numactl", msg.name, 7)) {
                        cout << "[NUMACTL] don't consider numactl process" << endl;
                        break;
                    }

                    if (!strncmp("time", msg.name, 4)) {
                        cout << "[TIME] don't consider time process" << endl;
                        break;
                    }

                    if (!strncmp("snumap-main", msg.name, 11)) {
                        cout << "[SNUMAP-MAIN] don't consider snumap-main process" << endl;
                        break;
                    }

                    scheduler->InsertParApp(msg.pid, msg.name);
                    scheduler->Schedule();

                    cout << "[BEGIN] pid: " << msg.pid << " name: " << msg.name << endl;

                    break;
                }
            case APP_END:
                {
                    app_end_data msg;
                    app_connection->ReadMSG(cmd, &msg);

#if APP_PROFILES
                    {
                        ParApp* app = scheduler->FindParAppByPID(msg.pid);
                        string app_profiles =
                            "log/"+to_string((long long int) msg.pid)+".profiles.csv";
                        ofstream dumpout(app_profiles.c_str());
                        app->DumpPerformanceProfiles(app_profiles);
                        dumpout.close();
                    }
#endif

                    /* delete the application from scheduler's list */
                    scheduler->DeleteParApp(msg.pid);

                    scheduler->SetSchedulingPhase(PROFILING_RUN);
                    scheduler->Schedule();

                    //scheduler->SetSchedulingPhase(GREEDY_ALLOCATION);

                    cout << "[END] pid: " << msg.pid << endl;

                    break;
                }
            default:
                {
                    break;
                }
        }

        app_connection->SendACK();
        app_connection->Close();
    }
}

int main(int argc, char *argv[])
{
#ifndef C3PO
    // Handle termination and error sginals
    struct sigaction sigact;
    sigact.sa_handler = (__sighandler_t)signal_handler;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_RESTART;
    sigaction(SIGINT, &sigact, NULL);
    sigaction(SIGTERM, &sigact, NULL);
    sigaction(SIGSEGV, &sigact, NULL);

    /*
     * build prototype
     */
    CPrototype* prototype = new CPrototype();

    /*
     * build architecture instance
     */
    CPUArchitecture* architecture =
        prototype->CreateArchitectureInstance();

    prototype->DetailedSystemInfo(architecture, NULL, NULL, "arch.txt");

    /*
     * Create scheduler instance and profiler
     */
#if defined(NUPOCO)
    printf("NuPoCo!\n");
    CoreAllocator* scheduler = prototype->CreateHeuristicsScheduler(architecture);
    CProfiler *profiler = prototype->CreateHeuristicsProfiler(architecture, scheduler);
#elif defined(GREEDY)
    CoreAllocator* scheduler = prototype->CreateGreedyScheduler(architecture);
    CProfiler *profiler = prototype->CreateHeuristicsProfiler(architecture, scheduler);
#elif defined(EQP)
    CoreAllocator* scheduler = prototype->CreateEqualPartitionScheduler(architecture);
    CProfiler *profiler = prototype->CreateSimpleProfiler(architecture, scheduler);
#else
    CoreAllocator* scheduler = prototype->CreateEqualPartitionScheduler(architecture);
    CProfiler *profiler = prototype->CreateSimpleProfiler(architecture, scheduler);
#endif

    scheduler->RegisterProfiler(profiler);

    /*
     * application connection setting
     */
    RTEAppConnection* app_connection =
        prototype->CreateApplicationSocketConnection();

    /*
     * prototype information
     */
    prototype->DetailedSystemInfo(architecture, scheduler, profiler,
            "log/detailed.system.info");

    system("touch ready");

    /*
     * run periodic scheduler
     */
    std::thread periodicScheduler(PeriodicSchedulerRoutine,
            scheduler,
            profiler,
            architecture);

    /*
     * run application communication listener
     */
    std::thread appListener(AppConnectionListener,
            app_connection,
            scheduler);

    /*
     * wait for termination
     */
    periodicScheduler.join();
    appListener.join();

    return 0;
#else
    // Handle termination and error sginals
    struct sigaction sigact;
    sigact.sa_handler = (__sighandler_t)signal_handler;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_RESTART;
    sigaction(SIGINT, &sigact, NULL);
    sigaction(SIGTERM, &sigact, NULL);
    sigaction(SIGSEGV, &sigact, NULL);

    /*
     * build prototype
     */
    CPrototype* prototype = new CPrototype();

    /*
     * build architecture instance
     */
    CPUArchitecture* architecture =
        prototype->CreateArchitectureInstance();

    prototype->DetailedSystemInfo(architecture, NULL, NULL, "arch.txt");

    /*
     * Set all cores to max freq
     */
    cout << "Set all cores to max_freq: " << power->get_max_freq() << " kHz\n";
    power->set_freq_all_cores(power->get_max_freq());

#ifdef INTEL
    /*
     * Set all MCT to default bandwidth
     */
    cout << "Set all MCT's to default power cap" << endl;
    for (size_t i = 0; i < architecture->num_memory_nodes(); ++i)
        power->rapl_throttle_dram(i, 1);
#endif /* INTEL */

    /*
     * scheduler setting
     */
    CoreAllocator* scheduler = prototype->CreateC3POScheduler(architecture);

    /*
     * profiler setting
     */
    CProfiler *profiler =
        prototype->CreateC3POProfiler(architecture, scheduler);

    scheduler->RegisterProfiler(profiler);

    /*
     * application connection setting
     */
    RTEAppConnection* app_connection =
        prototype->CreateApplicationSocketConnection();

    /*
     * prototype information
     */
    prototype->DetailedSystemInfo(architecture, scheduler, profiler,
            "log/detailed.system.info");

    system("touch ready");

    /*
     * run periodic scheduler
     */
    std::thread periodicScheduler(PeriodicSchedulerRoutine_C3PO,
            scheduler,
            profiler,
            architecture);

    /*
     * run application communication listener
     */
    std::thread appListener(AppConnectionListener,
            app_connection,
            scheduler);

    /*
     * wait for termination
     */
    periodicScheduler.join();
    appListener.join();

    power->restore_orig_state();
    return 0;
#endif
}
