//------------------------------------------------------------------------------
/// @brief PMU interface protocol
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription pmu_interface.cpp
/// PMU interface protocol implementation
/// 
//------------------------------------------------------------------------------

#include <cassert>
#include <cstdio>
#include <cstring>
#include <sys/ioctl.h>
#include <unistd.h>

#include "pmu_interface.h"

////////////////////////////////////////////////////////////////////////////////
//
// Implementation for AMD64 system
//
////////////////////////////////////////////////////////////////////////////////
#if defined(AMD)

#define TOTAL_CYCLES                0
#define TOTAL_INSTRUCTIONS          1
#define L2_CACHE_MISSES             1
#define STALL_CYCLES                2
#define LLC_MISSES                  3

#define LOCAL_DATA_READ_MISS        1
#define LOCAL_WRITE_MISS            2
#define REMOTE_DATA_READ_MISS       3
#define REMOTE_WRITE_MISS           4

#ifndef __NR_perf_event_open
#if defined(__PPC__)
#define __NR_perf_event_open	319
#elif defined(__i386__)
#define __NR_perf_event_open	336
#elif defined(__x86_64__)
#define __NR_perf_event_open	298
#elif defined(__tile__)
#define __NR_perf_event_open	241
#else
#error __NR_perf_event_open must be defined
#endif
#endif

#define PRINT_COUNT	0

struct read_format {
    unsigned long long value;         /* The value of the event */
    unsigned long long time_enabled;  /* if PERF_FORMAT_TOTAL_TIME_ENABLED */
};

PMUReqManager::PMUReqManager()
{
    memset(_fd, 0, sizeof(_fd)) ;
    InitPerfEventAttr();
}

PMUReqManager::~PMUReqManager()
{
}

void PMUReqManager::InitPerfEventAttr()
{
    memset(&_attr, 0, sizeof(_attr));

    for (int i = 0; i < MAX_EVENTS; i++) {
        // default attribute configuration
        _attr[i].inherit = 1;
        _attr[i].disabled = 1;
        _attr[i].size = sizeof(_attr[i]);
        _attr[i].read_format = PERF_FORMAT_TOTAL_TIME_ENABLED;
        _attr[i].type = TOTAL_CYCLES_TYPE;
        _attr[i].config = TOTAL_CYCLES_CONFIG;
    }

    for (int i = 0; i < MAX_EVENTS; i++) {
        // default event type
        _event_type[i] = PER_CORE;
    }
}

void PMUReqManager::InitPerfEventAttr(
        std::pair<__u32, __u64>* attr_mapping,
        event_type* attr_type,
        int num_events)
{
    if (num_events > MAX_EVENTS) {
        cout << "maximum number of events: " << MAX_EVENTS
            << " can't register " << num_events << " events." << endl;
        return;
    }

    for (int i = 0; i < num_events; i++) {
        // attribute configuration
        _attr[i].inherit = 1;
        _attr[i].disabled = 1;
        _attr[i].size = sizeof(_attr[i]);
        _attr[i].read_format = PERF_FORMAT_TOTAL_TIME_ENABLED;
        _attr[i].type = attr_mapping[i].first;
        _attr[i].config = attr_mapping[i].second;
    }

    for (int i = 0; i < num_events; i++) {
        // event type configuration
        _event_type[i] = attr_type[i];
    }
}

void PMUReqManager::GetCurrentCount(rte_event_values& val)
{
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    val = _cur;
}

void PMUReqManager::GetCurrentCount(rte_event_values& val,
        unsigned long* per_thread_llc_misses)
{
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    val = _cur;
}

void PMUReqManager::OpenEventFDs()
{
    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                _fd[j][i] = _fd[j][i - (i % NUM_CORES_IN_NODE)];
                continue;
            }

            // perf_event_open(perf_event_attr, pid, cpuid, is_group, flag);
            _fd[j][i] = syscall(__NR_perf_event_open, &_attr[j], -1, i, -1, 0);
            if (_fd[j][i] < 0) {
                perror("sys_perf_event_open");
                abort();
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_ENABLE);
        }
    }
}

void PMUReqManager::CloseEventFDs() {
    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                continue;
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_DISABLE);
            close(_fd[j][i]);
        }
    }
}

void PMUReqManager::OpenEventFDs(int subset_index)
{
    for (int j = subset_index; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                _fd[j][i] = _fd[j][i - (i % NUM_CORES_IN_NODE)];
                continue;
            }

            // perf_event_open(perf_event_attr, pid, cpuid, is_group, flag);
            _fd[j][i] = syscall(__NR_perf_event_open, &_attr[j], -1, i, -1, 0);
            if (_fd[j][i] < 0) {
                perror("sys_perf_event_open");
                abort();
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_ENABLE);
        }
    }
}

void PMUReqManager::OpenEventFDsByPID(pid_t pid)
{
    ////for (int j = subset_index; j < MAX_EVENTS; j++) {
    //for (int i = 0; i < NUM_CORES; i++) {

    //  if (_event_type[j] == PER_NODE &&
    //      (i % NUM_CORES_IN_NODE != 0)) {
    //    _fd[j][i] = _fd[j][i - (i % NUM_CORES_IN_NODE)];
    //    continue;
    //  }

    //  // perf_event_open(perf_event_attr, pid, cpuid, is_group, flag);
    //  _fd[j][i] = syscall(__NR_perf_event_open, &_attr[j], -1, i, -1, 0);
    //  if (_fd[j][i] < 0) {
    //    perror("sys_perf_event_open");
    //    abort();
    //  }

    //  ioctl(_fd[j][i], PERF_EVENT_IOC_ENABLE);
    //}
    ////}
}

void PMUReqManager::CloseEventFDs(int subset_index)
{
    for (int j = subset_index; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                continue;
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_DISABLE);
            close(_fd[j][i]);
        }
    }
}

void PMUReqManager::RecordAllCounters() {
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    ReadCounters(&_cur);
}

void PMUReqManager::ReadCounters(rte_event_values *values)
{
    struct read_format count_result;
    unsigned long long count = 0;

    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE && (i % NUM_CORES_IN_NODE != 0)) {
                values->values[j][i] = 0;
                //values->values[j][i] = values->values[j][i-(i % NUM_CORES_IN_NODE)];
                continue;
            }

            size_t res;
            res = read(_fd[j][i], &count_result, sizeof(struct read_format));
            if (res != sizeof(struct read_format)) {
                perror("read file descriptor error");
                abort();
            }

            values->values[j][i] = count_result.value;
            values->timestamp[j][i] = count_result.time_enabled;
#if PRINT_COUNT
            cout << "event: " << j << " cpu: " << i
                << " count: " << values->values[j][i] << endl;
#endif
        }
    }
}

#elif defined(INTEL)

PMUReqManager::PMUReqManager()
{
    InitPerfEventAttr();
}

PMUReqManager::~PMUReqManager()
{
}

void PMUReqManager::InitPerfEventAttr()
{
    cout << string(80, '~') << endl << "Setting Intel PCM up..." << endl << endl;

    // Useful to cought various signals. But, might not be needed all the time
    // set_signal_handlers();
    m = PCM::getInstance();

    EventSelectRegister def_event_select_reg;
    def_event_select_reg.value          = 0;
    def_event_select_reg.fields.usr     = 1;
    def_event_select_reg.fields.os      = 1;
    def_event_select_reg.fields.enable  = 1;

    PCM::ExtendedCustomCoreEventDescription conf;
    // number of events must be divisible by 4 (why?) TODO
    int _nGPCounters = 4;
    conf.fixedCfg = NULL; // default
    conf.nGPCounters = _nGPCounters;
    switch(m->getCPUModel())
    {
        case PCM::WESTMERE_EX:
            // OFFCORE_RESPONSE.ANY_REQUEST.LOCAL_DRAM
            conf.OffcoreResponseMsrValue[0] = 0x40FF;

            // OFFCORE_RESPONSE.ANY_REQUEST.REMOTE_DRAM
            conf.OffcoreResponseMsrValue[1] = 0x20FF;
            break;

        case PCM::JAKETOWN:
        case PCM::IVYTOWN:
            // OFFCORE_RESPONSE.*.LOCAL_DRAM
            conf.OffcoreResponseMsrValue[0] = 0x780400000 | 0x08FFF;

            // OFFCORE_RESPONSE.*.REMOTE_DRAM
            conf.OffcoreResponseMsrValue[1] = 0x7ff800000 | 0x08FFF;
            break;

        case PCM::HASWELLX:
            // OFFCORE_RESPONSE.*.LOCAL_DRAM
            conf.OffcoreResponseMsrValue[0] = 0x600400000 | 0x08FFF;

            // OFFCORE_RESPONSE.*.REMOTE_DRAM
            conf.OffcoreResponseMsrValue[1] = 0x67f800000 | 0x08FFF;
            break;
        default:
            std::cout << "[PCM] CPU Model not supported." << std::endl;
            exit(EXIT_FAILURE);
    }

    EventSelectRegister regs[_nGPCounters];
    conf.gpCounterCfg = regs;
    for(int i = 0; i < _nGPCounters; ++i)
        regs[i] = def_event_select_reg;

    regs[0].fields.event_select = 0xB7; // OFFCORE_RESPONSE 0 event
    regs[0].fields.umask = 0x01;
    regs[1].fields.event_select = 0xBB; // OFFCORE_RESPONSE 1 event
    regs[1].fields.umask = 0x01;

    // 0xA2 and 0x01 --> Resource-related stall cycles.
    // 0xA3 and 0x04 --> Total execution stalls.
    regs[2].fields.event_select = 0xA2; // RESOURCE_STALLS event
    regs[2].fields.umask = 0x01;

    // regs[3].fields.event_select = 0x0E;   // UOPS_ISSUED.ANY event
    // regs[3].fields.umask = 0x01;
    // regs[3].fields.any_thread = 1;
    // regs[3].fields.invert = 1;
    // regs[3].fields.cmask = 1;

    m->resetPMU();
    PCM::ErrorCode status = m->program(PCM::EXT_CUSTOM_CORE_EVENTS, &conf);
    switch (status)
    {
        case PCM::Success:
            break;
        case PCM::MSRAccessDenied:
            cerr << "Access to Intel(r) PCM has denied (no MSR or PCI CFG space \
                access)." << endl;
            exit(EXIT_FAILURE);
        case PCM::PMUBusy:
            cerr << "Access to Intel(r) PCM has denied (Performance Monitoring \
                Unit is occupied by other application). Try to stop the \
                application that uses PMU." << endl;
            cerr << "Alternatively you can try to reset PMU configuration at \
                your own risk. Try to reset? (y/n)" << endl;
            char yn;
            std::cin >> yn;
            if ('y' == yn)
            {
                m->resetPMU();
                cerr << "PMU configuration has been reset. Try to rerun the \
                    program again." << endl;
            }
            exit(EXIT_FAILURE);
        default:
            cerr << "Access to Intel(r) PCM has denied (Unknown error)." << endl;
            exit(EXIT_FAILURE);
    }

    if(!m->hasPCICFGUncore())
    {
        cerr << "Jaketown, Ivytown or Haswell Server CPU is required for this \
            tool!" << endl;
        if(m->memoryTrafficMetricsAvailable())
            cerr << "For processor-level memory bandwidth statistics please use \
                pcm.x" << endl;
        exit(EXIT_FAILURE);
    }

    /// XXX actually needed?
    // int num_sockets = 4;
    // if(m->getNumSockets() > num_sockets)
    // {
    //     cerr << "Only systems with up to "<<num_sockets<<" sockets are supported! Program aborted" << endl;
    //     exit(EXIT_FAILURE);
    // }

    cout << string(80, '~') << endl;
}

void PMUReqManager::GetCurrentCount(rte_event_values& val)
{
    //std::lock_guard<std::mutex>  lock(_cur_mutex);
    val = _cur;
}

void PMUReqManager::RecordAllCounters() {
    //std::lock_guard<std::mutex>  lock(_cur_mutex);
    ReadCounters(&_cur);
}

void PMUReqManager::ReadCounters(rte_event_values *values)
{
    m->getAllCounterStates(values->sstate, values->sktstate, values->cstate);
}

#elif defined(__tile__)
#define TOTAL_CYCLES                0
#define TOTAL_INSTRUCTIONS          1
#define L2_CACHE_MISSES             1
#define STALL_CYCLES                2
#define LLC_MISSES                  3

#define LOCAL_DATA_READ_MISS        1
#define LOCAL_WRITE_MISS            2
#define REMOTE_DATA_READ_MISS       3
#define REMOTE_WRITE_MISS           4

#ifndef __NR_perf_event_open
#if defined(__PPC__)
#define __NR_perf_event_open	319
#elif defined(__i386__)
#define __NR_perf_event_open	336
#elif defined(__x86_64__)
#define __NR_perf_event_open	298
#elif defined(__tile__)
#define __NR_perf_event_open	241
#else
#error __NR_perf_event_open must be defined
#endif
#endif

#define PRINT_COUNT	0

struct read_format {
    unsigned long long value;         /* The value of the event */
    unsigned long long time_enabled;  /* if PERF_FORMAT_TOTAL_TIME_ENABLED */
};

PMUReqManager::PMUReqManager()
{
    memset(_fd, 0, sizeof(_fd)) ;
    InitPerfEventAttr();
}

PMUReqManager::~PMUReqManager()
{
}

void PMUReqManager::InitPerfEventAttr()
{
    memset(&_attr, 0, sizeof(_attr));

    for (int i = 0; i < MAX_EVENTS; i++) {
        // default attribute configuration
        _attr[i].inherit = 1;
        _attr[i].disabled = 1;
        _attr[i].size = sizeof(_attr[i]);
        _attr[i].read_format = PERF_FORMAT_TOTAL_TIME_ENABLED;
        _attr[i].type = TOTAL_CYCLES_TYPE;
        _attr[i].config = TOTAL_CYCLES_CONFIG;
    }

    for (int i = 0; i < MAX_EVENTS; i++) {
        // default event type
        _event_type[i] = PER_CORE;
    }
}

void PMUReqManager::InitPerfEventAttr(
        std::pair<__u32, __u64>* attr_mapping,
        event_type* attr_type,
        int num_events)
{
    if (num_events > MAX_EVENTS) {
        cout << "maximum number of events: " << MAX_EVENTS
            << " can't register " << num_events << " events." << endl;
        return;
    }

    for (int i = 0; i < num_events; i++) {
        // attribute configuration
        _attr[i].inherit = 1;
        _attr[i].disabled = 1;
        _attr[i].size = sizeof(_attr[i]);
        _attr[i].read_format = PERF_FORMAT_TOTAL_TIME_ENABLED;
        _attr[i].type = attr_mapping[i].first;
        _attr[i].config = attr_mapping[i].second;
    }

    for (int i = 0; i < num_events; i++) {
        // event type configuration
        _event_type[i] = attr_type[i];
    }
}

void PMUReqManager::GetCurrentCount(rte_event_values& val)
{
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    val = _cur;
}

void PMUReqManager::GetCurrentCount(rte_event_values& val,
        unsigned long* per_thread_llc_misses)
{
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    val = _cur;
}

void PMUReqManager::OpenEventFDs()
{
    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                _fd[j][i] = _fd[j][i - (i % NUM_CORES_IN_NODE)];
                continue;
            }

            // perf_event_open(perf_event_attr, pid, cpuid, is_group, flag);
            _fd[j][i] = syscall(__NR_perf_event_open, &_attr[j], -1, i, -1, 0);
            if (_fd[j][i] < 0) {
                perror("sys_perf_event_open");
                abort();
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_ENABLE);
        }
    }
}

void PMUReqManager::CloseEventFDs() {
    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                continue;
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_DISABLE);
            close(_fd[j][i]);
        }
    }
}

void PMUReqManager::OpenEventFDs(int subset_index)
{
    for (int j = subset_index; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                _fd[j][i] = _fd[j][i - (i % NUM_CORES_IN_NODE)];
                continue;
            }

            // perf_event_open(perf_event_attr, pid, cpuid, is_group, flag);
            _fd[j][i] = syscall(__NR_perf_event_open, &_attr[j], -1, i, -1, 0);
            if (_fd[j][i] < 0) {
                perror("sys_perf_event_open");
                abort();
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_ENABLE);
        }
    }
}

void PMUReqManager::OpenEventFDsByPID(pid_t pid)
{
}

void PMUReqManager::CloseEventFDs(int subset_index)
{
    for (int j = subset_index; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE &&
                    (i % NUM_CORES_IN_NODE != 0)) {
                continue;
            }

            ioctl(_fd[j][i], PERF_EVENT_IOC_DISABLE);
            close(_fd[j][i]);
        }
    }
}

void PMUReqManager::RecordAllCounters() {
    std::lock_guard<std::mutex>  lock(_cur_mutex);
    ReadCounters(&_cur);
}

void PMUReqManager::ReadCounters(rte_event_values *values)
{
    struct read_format count_result;
    unsigned long long count = 0;

    for (int j = 0; j < MAX_EVENTS; j++) {
        for (int i = 0; i < NUM_CORES; i++) {
            if (_event_type[j] == PER_NODE && (i % NUM_CORES_IN_NODE != 0)) {
                values->values[j][i] = 0;
                //values->values[j][i] = values->values[j][i-(i % NUM_CORES_IN_NODE)];
                continue;
            }

            size_t res;
            res = read(_fd[j][i], &count_result, sizeof(struct read_format));
            if (res != sizeof(struct read_format)) {
                perror("read file descriptor error");
                abort();
            }

            values->values[j][i] = count_result.value;
            values->timestamp[j][i] = count_result.time_enabled;
#if PRINT_COUNT
            cout << "event: " << j << " cpu: " << i
                << " count: " << values->values[j][i] << endl;
#endif
        }
    }
}

#endif
