//------------------------------------------------------------------------------
/// @brief PMU interface protocol
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
/// 
/// @decription pmu_interface.h
/// PMU interface protocol description.
/// 
//------------------------------------------------------------------------------

#ifndef __PMU_INTERFACE_H_
#define __PMU_INTERFACE_H_

#include <cstdlib>
#include <unistd.h>
#include <linux/perf_event.h>
#include <sys/epoll.h>
#include <mutex>

#include <tuple>
#include "common.h"
#include "rte_event.h"
#include "rte_cpu.h"
#include "scheduler.h"

#if defined(AMD)

//------------------------------------------------------------------------------
/// @brief RTE-PMU Request Manager
///
class PMUReqManager {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief default constructor
    PMUReqManager();

    /// @brief default destructor
    ~PMUReqManager();

    /// @}

    /// @name PMU request manager attributes
    /// @{

    /// @brief perf event attributes initial setting
    void InitPerfEventAttr();

    /// @brief perf event attributes setting from input
    ///
    /// @param attribute mapping information
    /// @param attribute type information
    /// @param the number of registers to be registered
    void InitPerfEventAttr(
        std::pair<__u32, __u64>* attr_mapping,
        event_type* attr_type,
        int num_events);

    /// @brief get current count values
    ///
    /// @param pointer to save current count values
    void GetCurrentCount(rte_event_values& val);
    void GetCurrentCount(rte_event_values& val,
        unsigned long* values);

    /// @brief open event file descriptors
    void OpenEventFDs();

    /// @brief open a subset of event file descriptors
    ///
    /// @param starting event index
    /// @description this open event file descriptors from the index to last
    void OpenEventFDs(int subset_index);
    void OpenEventFDsByPID(pid_t pid);

    /// @brief close event file descriptors
    void CloseEventFDs();

    /// @brief close a subset of event file descriptors
    ///
    /// @param starting event index
    /// @description this close event file descriptors from the index to last
    void CloseEventFDs(int subset_index);

    /// @brief record count values int ocount values
    void RecordAllCounters();

    /// @}

  private:
    /// @name PMU request manager private attributes
    /// @{

    /// @brief read count values from file descriptors
    void ReadCounters(rte_event_values *values);

    /// @}

    std::list<int> _fd_threads;
    int _fd[MAX_EVENTS][NUM_CORES]; ///< event-perf file descriptors
    struct perf_event_attr _attr[MAX_EVENTS]; ///< event-perf attributes
    event_type _event_type[MAX_EVENTS]; ///< event type information
    rte_event_values _cur; ///< recently stored count values
    std::mutex _cur_mutex;
};

#elif defined(INTEL)
class PMUReqManager {
  public:
    PMUReqManager();
    ~PMUReqManager();

    void InitPerfEventAttr();

    void GetCurrentCount(rte_event_values& val);

    void RecordAllCounters();

  private:
    void ReadCounters(rte_event_values *values);

    rte_event_values _cur; ///< recently stored count values
    
    PCM* m;
};

#elif defined(INTEL)

class PMUReqManager {
  public:
    PMUReqManager();
    ~PMUReqManager();

    void InitPerfEventAttr();
    void InitPerfEventAttr(
        std::pair<__u32, __u64>* attr_mapping,
        event_type* attr_type,
        int num_events);

    void GetCurrentCount(rte_event_values& val);
    void GetCurrentCount(rte_event_values& val,
        unsigned long* values);

    void OpenEventFDs();

    void OpenEventFDs(int subset_index);
    void OpenEventFDsByPID(pid_t pid);

    void CloseEventFDs();

    void CloseEventFDs(int subset_index);

    void RecordAllCounters();

  private:
    void ReadCounters(rte_event_values *values);

    std::list<int> _fd_threads;
    int _fd[MAX_EVENTS][NUM_CORES]; ///< event-perf file descriptors
    struct perf_event_attr _attr[MAX_EVENTS]; ///< event-perf attributes
    event_type _event_type[MAX_EVENTS]; ///< event type information
    rte_event_values _cur; ///< recently stored count values
    std::mutex _cur_mutex;
};

#endif

#endif

