//------------------------------------------------------------------------------
/// @brief ARM-RTE Power Managment API
/// @author Camilo Celis Guzman <camilo@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @description power_interface.h
/// This is an API to manage power via the cpufreq subsystem.
/// It makes I/O calls to sys filesystem.
///
//------------------------------------------------------------------------------

#ifndef __POWER_INTERFACE_H___
#define __POWER_INTERFACE_H___

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cassert>
#include <iomanip>    //< pretty print
#include <vector>
#include <numa.h>     //< numa_num_task_nodes()
#include <unistd.h>   // usleep()

#include "common.h"
#include "misc.h"

// TODO: Detect if userpsace governor is supported
/* #define GOV_PERF */
#define GOV_USRS

#ifdef GOV_PERF
#define DEF_GOVERNOR "performance"
#else
#define DEF_GOVERNOR "userspace"
#endif

using namespace std;

typedef unsigned int freq_t;
typedef unsigned int watt_t;

//------------------------------------------------------------------------------
/// @brief power management API
///
class PowerInterface{
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking no arguments
    PowerInterface();

    /// @brief default destructor
    ~PowerInterface();

    /// @}

    /// @brief restore original freq and governor values to all cores
    void restore_orig_state();

    /// @name Power Capping
    /// @{

    /// @brief gets the system power cap in Watts
    watt_t get_powercap(void) { return power_cap_; }

    /// @brief sets the per node power cap value used by other functions
    /// @param powercap new powercap value in Watts
    void set_per_node_powercap(watt_t powercap) { power_cap_ = powercap; }

    /// @}

    /// @name Frequency related API
    /// @{

    /// @brief sets the frequency of a core
    /// @param core_id  core to set the frequency of
    /// @param freq     frequency to set the core to
    /// @remarks
    ///    the frequency most be in the range of allowed frequencies
    void set_freq(int core_id, int freq);

    /// @brief sets the frequency of all the cores
    /// @param freq frequency to set all the cores to
    /// @remarks
    ///   the frequency most be in the range of allowed frequencies
    void set_freq_all_cores(int freq);

    /// @brief sets the frequency of some cores
    /// @param cores  vector of cores ids to set their frequency
    /// @param freq   frequency to set all the cores to
    /// @remarks
    ///   the frequency most be in the range of allowed frequencies
    void set_freq_some_cores(const vector<int>& cores, int freq);

    /// @brief Updates and returns the frequency of a given core
    /// @param core_id  core id to update the frequency of
    /// @return frequency of core identified by core_id
    /// @remarks
    ///   asserts if the core_id is within the available cores
    int get_freq(int core_id);

    /// @brief Updates and returns the frequency of a given core
    /// @param core_id  core id to update the frequency of
    /// @return frequency of core identified by core_id
    static int GetFreq(int core_id);

    /// @brief Updates the frequency of all the cores
    /// @return vector with the frequencies of all cores
    vector<int> get_freq_all_cores();

    /// @brief Returns max frequency for given core id
    /// @param core_id core id
    int get_max_freq(int core_id = 0) { return range_freqs_[core_id].front(); }

    /// @brief Returns min frequency for given core id
    /// @param core_id core id
    int get_min_freq(int core_id = 0) { return range_freqs_[core_id].back(); }

    /// @brief Returns a vector of frequencies that core_id can be set to
    /// @param core_id core to query its frequencies range
    /// @return vector of possible scaling frequencies
    /// @remarks
    ///    Note that this is just a retrieval of an array of ranges as
    ///    the frequencies are assumed to be static
    vector<int> get_freq_range(int core_id = 0);

    /// @brief Returns the transational latency to transition from one freq to other
    /// @param core_id core id
    int get_freq_latency(int core_id = 0);

    /// @}

    /// @name Governor related API
    /// @{

    /// @brief returns governor used for all cores
    string get_governor() { return governor_; }

    /// @brief returns available governors (for core 0)
    vector<string> get_available_governors() { return available_governors_; }

    /// @}


    /// @name RAPL related API
    /// @{

    // TODO: DRAM is not necessarily a subpackage of CPU packages
    /// @brief Returns the current energy consumtion for package package_id in J
    /// @param package_id package id
    /// @note seems that the package name and id are inverted! XXX
    static double rapl_get_package_energy(int package_id);

    /// @brief Returns the current energy consumtion for DRAM in J
    /// @param package_id package id
    /// @note seems that the package name and id are inverted! XXX
    static double rapl_get_dram_energy(int package_id);

    /// @brief Returns the package name
    /// @param package_id package id
    string rapl_get_package_name(int package_id);

    /// @brief Set per-node RAPL powercap
    string rapl_set_per_node_powercap(int package_id);

    /// @brief Enable DRAM power capping constrains to all DRAM controllers
    string rapl_enable_dram_constraint();

    /// @brief Disable DRAM power capping constrains to all DRAM controllers
    string rapl_disable_dram_constraint();

    /// @brief Enable CPU capping constrains to all sockets controllers
    string rapl_enable_cpu_constraint();

    /// @brief Disable CPU power capping constrains to all sockets controllers
    string rapl_disable_cpu_constraint();

    /// @brief Change DRAM controllers time window
    /// @param time_us new time window in us
    string rapl_change_dram_constraint_time_window(unsigned int time_us);

    /// @brief Throttle DRAM by a ratio of its max. power consumption estimate
    /// @param package_id DRAM controller package id
    /// @param ratio ratio to throttle the DRAM power by
    string rapl_throttle_dram(int package_id, float ratio);

    /// @brief Restores the power limit to 0 (no power limit) and disables
    /// power capping (as default)
    void rapl_restore_dram_power_limit();

    /// @brief Restores the power limit to 0 (no power limit) and disables
    /// power capping (as default)
    void rapl_restore_cpu_power_limit();

    /// @}

    /// @name
    /// @{ Pretty print functions

    /// @brief Pretty prints frequencies and cores ids for all cores
    void pp_freq_all_cores();

    /// @brief Pretty prints frequencies and cores ids for some cores
    /// @param cores vector of cores' ids to pretty print
    void pp_freq_some_cores(const vector<int>& cores);

    /// @}

    /// @name
    /// @{ Power Gating Cores

    /// @brief Power gate a core
    /// @param core id
    static void power_gate_core(int core_id);


    /// @brief Power gate a set of cores
    /// @param cores id embedded in a vector
    static void power_gate_cores(vector<int> cores);

    /// @brief Activates all cores in case they had been power gated
    void activate_all_cores(void);

    /// @}

    /// @name
    /// @{ Miscellaneous

    /// @brief Return number of NUMA nodes (uses numa.h)
    static size_t get_num_numa_nodes() { return numa_num_task_nodes(); }

    /// @}
  private:
    /// @brief initializes the frequencies of all cores, and updates
    ///      the ranges of frequencies for each core.
    ///      It also stores the original states of freqs and governors
    void _init_state();

    /// @brief Sets the scaling governor for all the cores in the system
    /// @param governor string containing the governor name.
    /// @remarks
    //    Note that if the governor name is incorrect then the old
    //    governor is maintained as current governor
    void _set_governor(string governor);

    /// @{ Intel RAPL specific methods

    /// @brief checks if we have access to Intel RAPL at root and package level
    /// @remakrs
    //    Returns True if we have RAPL, False otherwise
    bool _check_intel_rapl(void);

    /// @brief cahnges the DRAM constrain value (enaled)
    string _rapl_change_dram_constraint(string new_value, string field_to_change);

    /// @brief Enable system-wide RAPL powercap
    string _rapl_change_cpu_constraint(string new_value, string field_to_change);

    /// @}

    /// @{ sysfs specific methods

    /// @brief write to a sysfs file
    /// @param path sysfs file absolute path
    /// @param data data to write to the sysfs file
    /// @remaerks writes to stderr in case of error
    static void _write_sysfs(const string& path, const string& data);

    /// @brief reads an int from a sysfs file
    /// @param path sysfs file absolute path
    /// @return integer value from sysfs file
    /// @remaerks writes to stderr in case of error
    static int _read_sysfs(const string& path);

    /// @brief reads a string from a sysfs file
    /// @param path sysfs file absolute path
    /// @return integer value from sysfs file
    /// @remaerks writes to stderr in case of error
    static string _read_str_sysfs(const string& path);

    /// @}

    /// @brief returns the number of present cores
    static int _get_num_cores();


    const int num_cores_;                                 //< detected # cores
    const int num_nodes_;                                 //< detected # nodes
    const string prefix_;                                 //< sysfs base path
    bool has_rapl_;                                       //< is RAPL enabled?
    vector<string> available_governors_;                  //< vector of governors
    string governor_;                                     //< current governor
    vector< vector<int> > range_freqs_;                   //< range of freqs

    vector<int> orig_max_freqs;                           //< original max freqs
    vector<int> orig_min_freqs;                           //< original min freqs
    vector<string> orig_governors;                        //< original governors

    watt_t power_cap_;                                    //< powercap in watts
};

#endif
