//------------------------------------------------------------------------------
/// @brief ARM - RTE online performance profiler
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
//------------------------------------------------------------------------------

#include "profiler.h"

//------------------------------------------------------------------------------
// CSimpleProfiler
CSimpleProfiler::CSimpleProfiler(CPUArchitecture* arch, CoreAllocator* scheduler)
    : arch_(arch), scheduler_(scheduler)
{
#if defined(AMD)
    pmu_manager_ = new PMUReqManager();

    int num_events = 4;
    attr_mapping_ = (std::pair<__u32, __u64>*)
        malloc(num_events * sizeof(std::pair<__u32, __u64>));

    attr_mapping_[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    attr_mapping_[1] = make_pair<__u32, __u64>
        (TOTAL_INSTRUCTIONS_TYPE, TOTAL_INSTRUCTIONS_CONFIG);
    attr_mapping_[2] = make_pair<__u32, __u64>
        (STALL_CYCLES_TYPE, STALL_CYCLES_CONFIG);
    attr_mapping_[3] = make_pair<__u32, __u64>
        (LLC_MISSES_TYPE, LLC_MISSES_CONFIG);

    attr_type_ = (event_type*)malloc(num_events * sizeof(event_type));
    attr_type_[0] = PER_CORE;
    attr_type_[1] = PER_CORE;
    attr_type_[2] = PER_CORE;
    attr_type_[3] = PER_NODE;

    pmu_manager_->InitPerfEventAttr(attr_mapping_, attr_type_, num_events);
    pmu_manager_->OpenEventFDs();
#elif defined(INTEL)
    pmu_manager_ = new PMUReqManager();
    pmu_manager_->RecordAllCounters();

#elif defined(__tile__)
    pmu_manager_ = new PMUReqManager();

    int num_events = 5;
    attr_mapping_ = (pair<__u32, __u64>*)
        malloc(num_events * sizeof(pair<__u32, __u64>));

    attr_mapping_[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    attr_mapping_[1] = make_pair<__u32, __u64>
        (LOCAL_DATA_READ_MISS_TYPE, LOCAL_DATA_READ_MISS_CONFIG);
    attr_mapping_[2] = make_pair<__u32, __u64>
        (LOCAL_WRITE_MISS_TYPE, LOCAL_WRITE_MISS_CONFIG);
    attr_mapping_[3] = make_pair<__u32, __u64>
        (REMOTE_DATA_READ_MISS_TYPE, REMOTE_DATA_READ_MISS_CONFIG);
    attr_mapping_[4] = make_pair<__u32, __u64>
        (REMOTE_WRITE_MISS_TYPE, REMOTE_WRITE_MISS_CONFIG);

    attr_type_ = (event_type*)malloc(num_events * sizeof(event_type));
    attr_type_[0] = PER_CORE;
    attr_type_[1] = PER_CORE;
    attr_type_[2] = PER_CORE;
    attr_type_[3] = PER_CORE;
    attr_type_[4] = PER_CORE;

    pmu_manager_->InitPerfEventAttr(attr_mapping_, attr_type_, num_events);
    pmu_manager_->OpenEventFDs();
#endif
    cout << "Hello! I'm a Simpleprofiler" << endl;
}

CSimpleProfiler::~CSimpleProfiler()
{
    cout << "Bye~ from the profiler" << endl;
#if defined(AMD)
    free(attr_mapping_);
    free(attr_type_);
    pmu_manager_->CloseEventFDs();
#elif defined(INTEL)
    delete pmu_manager_;
#elif defined(__tile__)
    free(attr_mapping_);
    free(attr_type_);
    pmu_manager_->CloseEventFDs();
#endif
}

void CSimpleProfiler::ProfilerInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "memory contention profiler info. " << endl;

    out.close();

    return;
}

void CSimpleProfiler::StartProfiling()
{
#if defined(AMD)
    pmu_manager_->GetCurrentCount(start_);
#elif defined(INTEL)
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(start_);
#elif defined(__tile__)
    pmu_manager_->GetCurrentCount(start_);
#endif

    clock_gettime(CLOCK_REALTIME, &start);
}

void CSimpleProfiler::StopProfiling()
{
#if defined(AMD)
    const int TOTAL_CYCLES = 0;
    const int LLC_MISSES = 3;
    const int tiles_per_cluster = arch_->GetCPUNode(0)->num_tiles();

    rte_event_values end_;

    pmu_manager_->RecordAllCounters();

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (end_.values[0][0] == start_.values[0][0]) return;

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();

    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;
        int tiles = app->GetNumTiles();

        if (app->GetParallelRegion()) {
            list<CPUTile*> tile_list = app->GetTileSet();

            unsigned long sum_llc_misses = 0;
            unsigned long sum_total_cycles = 0;

            for (list<CPUTile*>::iterator tile_it = tile_list.begin();
                    tile_it != tile_list.end();
                    tile_it++) {
                CPUTile* tile = *tile_it;
                int tile_id = tile->id();
                unsigned long total_cycles = end_.values[TOTAL_CYCLES][tile_id]
                    - start_.values[TOTAL_CYCLES][tile_id];
                unsigned long llc_misses = 0.0;

                if (attr_type_[LLC_MISSES] == PER_NODE) {
                    int delegate_tile = tile_id - tile_id % tiles_per_cluster;
                    llc_misses =
                        (end_.values[LLC_MISSES][delegate_tile] -
                         start_.values[LLC_MISSES][delegate_tile]);
                    llc_misses = llc_misses / tiles_per_cluster;
                } else {
                    llc_misses =
                        (end_.values[LLC_MISSES][tile_id]
                         - start_.values[LLC_MISSES][tile_id]);
                }

                sum_llc_misses += llc_misses;
                sum_total_cycles += total_cycles;
            }

            app->UpdateLLCMissRate(sum_llc_misses, sum_total_cycles);
            std::cout << "app: " << app->name()
                << " llc_miss_rate: " << (double)sum_llc_misses/(double)sum_total_cycles
                << std::endl;
        }
    }
#elif defined(INTEL)
    rte_event_values end_;

    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(end_);

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();

    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;
        int tiles = app->GetNumTiles();

        if (app->GetParallelRegion()) {
            list<CPUTile*> tile_list = app->GetTileSet();

            unsigned long sum_llc_misses = 0;
            unsigned long sum_total_cycles = 0;

            for (list<CPUTile*>::iterator tile_it = tile_list.begin();
                    tile_it != tile_list.end();
                    tile_it++) {
                CPUTile* tile = *tile_it;
                int tile_id = tile->id();

                unsigned long total_cycles = getRefCycles(start_.cstate[tile_id], end_.cstate[tile_id]);
                // in Intel, LLC misses and DRAM accesses are collected at per-core level
                unsigned long llc_misses = getL3CacheMisses(start_.cstate[tile_id], end_.cstate[tile_id]);

                sum_llc_misses += llc_misses;
                sum_total_cycles += total_cycles;
            }

            app->UpdateLLCMissRate(sum_llc_misses, sum_total_cycles);
            std::cout << "app: " << app->name()
                << " llc_miss_rate: " << (double)sum_llc_misses/(double)sum_total_cycles
                << std::endl;
        }
    }

#elif defined(__tile__)
    const int TOTAL_CYCLES = 0;
    const int LOCAL_DATA_READ_MISS = 1;
    const int LOCAL_WRITE_MISS = 2;
    const int REMOTE_DATA_READ_MISS = 3;
    const int REMOTE_WRITE_MISS = 4;
    rte_event_values end_;

    pmu_manager_->RecordAllCounters();

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (end_.values[0][0] == start_.values[0][0]) return;

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();

    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;
        int tiles = app->GetNumTiles();

        if (app->GetParallelRegion()) {
            list<CPUTile*> tile_list = app->GetTileSet();

            unsigned long sum_llc_misses = 0;
            unsigned long sum_total_cycles = 0;

            for (list<CPUTile*>::iterator tile_it = tile_list.begin();
                    tile_it != tile_list.end();
                    tile_it++) {
                CPUTile* tile = *tile_it;
                int tile_id = tile->id();
                unsigned long total_cycles = end_.values[TOTAL_CYCLES][tile_id]
                    - start_.values[TOTAL_CYCLES][tile_id];
                unsigned long llc_misses = 0.0;

                llc_misses =
                    (end_.values[LOCAL_DATA_READ_MISS][tile_id]
                     - start_.values[LOCAL_DATA_READ_MISS][tile_id]);
                llc_misses +=
                    (end_.values[LOCAL_WRITE_MISS][tile_id]
                     - start_.values[LOCAL_WRITE_MISS][tile_id]);
                llc_misses +=
                    (end_.values[REMOTE_DATA_READ_MISS][tile_id]
                     - start_.values[REMOTE_DATA_READ_MISS][tile_id]);
                llc_misses +=
                    (end_.values[REMOTE_WRITE_MISS][tile_id]
                     - start_.values[REMOTE_WRITE_MISS][tile_id]);

                sum_llc_misses += llc_misses;
                sum_total_cycles += total_cycles;
            }

            app->UpdateLLCMissRate(sum_llc_misses, sum_total_cycles);
        }
    }
#endif
}

CHeuristicsProfiler::CHeuristicsProfiler(CPUArchitecture* arch,
        CoreAllocator* scheduler)
    : arch_(arch), scheduler_(scheduler)
{
#if defined(AMD)
    pmu_manager_ = new PMUReqManager();

    int num_events = 6;
    attr_mapping_ = (std::pair<__u32, __u64>*)
        malloc(num_events * sizeof(std::pair<__u32, __u64>));

    attr_mapping_[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    attr_mapping_[1] = make_pair<__u32, __u64>
        (STALL_CYCLES_TYPE, STALL_CYCLES_CONFIG);
    attr_mapping_[2] = make_pair<__u32, __u64>
        (LLC_ACCESSES_TYPE, LLC_ACCESSES_CONFIG);
    attr_mapping_[3] = make_pair<__u32, __u64>
        (LLC_MISSES_TYPE, LLC_MISSES_CONFIG);
    attr_mapping_[4] = make_pair<__u32, __u64>
        (MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_TYPE,
         MEMORY_CONTROLLER_REQUESTS_READ_REQUESTS_TO_DCT_CONFIG);
    attr_mapping_[5] = make_pair<__u32, __u64>
        (MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_TYPE,
         MEMORY_CONTROLLER_REQUESTS_WRITE_REQUESTS_TO_DCT_CONFIG);

    attr_type_ = (event_type*)malloc(num_events * sizeof(event_type));
    attr_type_[0] = PER_CORE;
    attr_type_[1] = PER_CORE;
    attr_type_[2] = PER_NODE;
    attr_type_[3] = PER_NODE;
    attr_type_[4] = PER_NODE;
    attr_type_[5] = PER_NODE;

    pmu_manager_->InitPerfEventAttr(attr_mapping_, attr_type_, num_events);
    pmu_manager_->OpenEventFDs();

    profiling_phase_ = NO_PROFILING;

#elif defined(INTEL)
    pmu_manager_ = new PMUReqManager();
    pmu_manager_->RecordAllCounters();

#endif

    cout << "Hello! I'm a profiler" << endl;
}

CHeuristicsProfiler::~CHeuristicsProfiler()
{
    cout << "Bye~ from the profiler" << endl;

#if defined(AMD)
    free(attr_mapping_);
    free(attr_type_);
    pmu_manager_->CloseEventFDs();
#elif defined(INTEL)
    delete pmu_manager_;
#endif
}

void CHeuristicsProfiler::ProfilerInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "memory contention profiler info. " << endl;

    out.close();

    return;
}

void CHeuristicsProfiler::StartProfiling()
{
#if defined(AMD)
    pmu_manager_->GetCurrentCount(start_);
#elif defined(INTEL)
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(start_);
#endif

    clock_gettime(CLOCK_REALTIME, &start);
}

void CHeuristicsProfiler::StopProfiling()
{
#if defined(AMD)
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(end_);

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (end_.values[0][0] == start_.values[0][0]) return;

    if (profiling_phase_ == NO_PROFILING) {

    } else if (profiling_phase_ == CLEAN_PROFILING) {
        std::cout << "profiling phase: " << profiling_phase_
            << " CleanProfiling" << std::endl;
        CleanProfiling();
    } else if (profiling_phase_ == MIXED_PROFILING) {
        std::cout << "profiling phase: " << profiling_phase_
            << " MixedProfiling" << std::endl;
        MixedProfiling();
    }
#elif defined(INTEL)
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(end_);

    if (profiling_phase_ == CLEAN_PROFILING) {
        std::cout << "profiling phase: " << profiling_phase_
            << " CleanProfiling" << std::endl;
        CleanProfiling();
    } else if (profiling_phase_ == MIXED_PROFILING) {
        std::cout << "profiling phase: " << profiling_phase_
            << " MixedProfiling" << std::endl;
        MixedProfiling();
    }
#endif
}

void CHeuristicsProfiler::CleanProfiling()
{
#if defined(AMD)
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
    elapsed_time /= 1000; // microseconds

    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    const int TOTAL_CYCLES = 0;
    const int STALL_CYCLES = 1;
    const int LLC_ACCESSES = 2;
    const int LLC_MISSES = 3;
    const int MCT_READS = 4;
    const int MCT_WRITES = 5;
    const int NUM_NODES = NUM_CORES / NUM_CORES_IN_NODE;

#ifdef DEBUG_MODE
    std::cout << "This is clean profiling" << std::endl;
#endif

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();
    for (map<pid_t, ParApp*>::iterator it = apps.begin();
         it != apps.end();
         it++) {
        ParApp* app = it->second;

        if (app->GetNeedProfiling()) {
#ifdef DEBUG_MODE
            std::cout << "get profiling" << std::endl;
            CPUNode* cpu_node = app->GetProfilingNode();
            if (cpu_node == NULL) {
                std::cout << "get profiling node: NULL" << std::endl;
                //continue;
            } else {
                std::cout << "get profiling node: " << app->GetProfilingNode()->id() << std::endl;
            }
#endif

            unsigned long total_cycles = 0;
            unsigned long stall_cycles = 0;
            unsigned long llc_accesses = 0;
            unsigned long llc_misses = 0;
            unsigned long mct_read_requests[NUM_NODES] = { 0, };
            unsigned long mct_write_requests[NUM_NODES] = { 0, };
            //unsigned long mct_requests[NUM_NODES] = { 0, };
            unsigned long average_mct_requests[NUM_NODES] = { 0, };

            // assuming profiled application occupies all cpu resources
            for (int i = 0; i < NUM_CORES; i+= NUM_CORES_IN_NODE) {
                for (int j = i; j < i+NUM_CORES_IN_NODE; j++) {
                    total_cycles += end_.values[TOTAL_CYCLES][j] -
                        start_.values[TOTAL_CYCLES][j];
                    stall_cycles += end_.values[STALL_CYCLES][j] -
                        start_.values[STALL_CYCLES][j];
                }

                llc_accesses += end_.values[LLC_ACCESSES][i] - start_.values[LLC_ACCESSES][i];
                llc_misses += end_.values[LLC_MISSES][i] - start_.values[LLC_MISSES][i];

                int node_num = i/NUM_CORES_IN_NODE;
                mct_read_requests[node_num] = end_.values[MCT_READS][i] - start_.values[MCT_READS][i];
                mct_write_requests[node_num] = end_.values[MCT_WRITES][i] - start_.values[MCT_WRITES][i];
            }

            total_cycles = total_cycles / NUM_CORES;
            stall_cycles = stall_cycles / NUM_CORES;
            llc_accesses = llc_accesses / NUM_CORES;
            llc_misses = llc_misses / NUM_CORES;

            unsigned long sum_mct_requests = 0;
            for (int i = 0; i < NUM_NODES; i++) {
                average_mct_requests[i] = (mct_read_requests[i] + mct_write_requests[i])/ NUM_CORES;
                sum_mct_requests += average_mct_requests[i];
            }

#ifdef DEBUG_MODE
            std::cout << "app: " << app->name()
                << " max_cycles: " << max_cycles
                << " total_cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " llc_accesses: " << llc_accesses
                << " llc_misses: " << llc_misses
                << " sum_mct_requests: " << sum_mct_requests << std::endl;
            std::cout << "mct_requests: ";
            for (int i = 0; i < NUM_NODES; i++)
                std::cout << " node" << i << " " << average_mct_requests[i];
            std::cout << std::endl;
#endif

            app->UpdateLLCAccessRate(llc_accesses, total_cycles - stall_cycles);
            app->UpdateLLCMissRate(llc_misses, total_cycles - stall_cycles);
            app->UpdateDRAMReqRate(sum_mct_requests, total_cycles - stall_cycles);
            app->UpdateLLCMissAccessRate(llc_misses, llc_accesses);

            app->UnsetProfiling();
            app->UnsetProfilingNode();

            app->SetRegionProfiled();
        }
    }
#elif defined(INTEL)
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
    elapsed_time /= 1000; // microseconds

    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();
    for (map<pid_t, ParApp*>::iterator it = apps.begin();
         it != apps.end();
         it++) {
        ParApp* app = it->second;

        if (app->GetNeedProfiling()) {
#ifdef DEBUG_MODE
            CPUNode* cpu_node = app->GetProfilingNode();
            if (cpu_node == NULL) {
                std::cout << "get profiling node: NULL" << std::endl;
                //continue;
            } else {
                std::cout << "get profiling node: " << app->GetProfilingNode()->id() << std::endl;
            }
#endif

            unsigned long total_cycles = 0;
            unsigned long stall_cycles = 0;
            unsigned long llc_accesses = 0;
            unsigned long llc_misses = 0;
            unsigned long mct_read_requests[NUM_NODES] = { 0, };
            unsigned long mct_write_requests[NUM_NODES] = { 0, };
            unsigned long mct_requests[NUM_NODES] = { 0, };

            // assuming profiled application occupies all cpu resources
            for (int i = 0; i < NUM_CORES; i+=NUM_CORES_IN_NODE) {
                for (int j = i; j < NUM_CORES_IN_NODE+i; j++) {
                    total_cycles += getRefCycles(start_.cstate[j], end_.cstate[j]);
                    stall_cycles = getNumberOfCustomEvents(2, start_.cstate[j], end_.cstate[j]);
                    // in Intel, LLC misses and DRAM accesses are collected at per-core level
                    llc_misses += getL3CacheMisses(start_.cstate[j], end_.cstate[j]);
                    llc_accesses += getL3CacheHits(start_.cstate[j], end_.cstate[j]) + getL3CacheMisses(start_.cstate[j], end_.cstate[j]);
                }

                int node_num = i/NUM_CORES_IN_NODE;
                // in Intel, we can't distinguish prefetches from reads for now.
                mct_read_requests[node_num] = getBytesReadFromMC(start_.sktstate[node_num], end_.sktstate[node_num])/64;
                mct_write_requests[node_num] = getBytesWrittenToMC(start_.sktstate[node_num], end_.sktstate[node_num])/64;
            }

            for (int i = 0; i < NUM_NODES; i++) {
                mct_requests[i] = mct_read_requests[i] + mct_write_requests[i];
            }

            total_cycles = total_cycles / NUM_CORES; // average per-core total cycles
            stall_cycles = stall_cycles / NUM_CORES; // average per-core stall cycles
            llc_misses = llc_misses / NUM_CORES; // average per-core llc misses
            llc_accesses = llc_accesses / NUM_CORES; // average per-core llc accesses

            unsigned long average_mct_requests[NUM_NODES] = { 0, };
            unsigned long sum_mct_requests = 0;
            for (int i = 0; i < NUM_NODES; i++) {
                average_mct_requests[i] = mct_requests[i] / NUM_CORES;
                sum_mct_requests += average_mct_requests[i];
            }

#ifdef DEBUG_MODE
            std::cout << "[clean_profiling] app_region: " << app->GetParallelRegion() << std::endl;

            std::cout << "app: " << app->name()
                << " max_cycles: " << max_cycles
                << " total_cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " llc_accesses: " << llc_accesses
                << " llc_misses: " << llc_misses
                << " sum_mct_requests: " << sum_mct_requests << std::endl;
            std::cout << "mct_requests: ";
            for (int i = 0; i < NUM_NODES; i++)
                std::cout << " node" << i << " " << average_mct_requests[i];
            std::cout << std::endl;
#endif

            app->UpdateLLCAccessRate(llc_accesses, total_cycles - stall_cycles);
            app->UpdateLLCMissRate(llc_misses, total_cycles - stall_cycles);
            app->UpdateDRAMReqRate(sum_mct_requests, total_cycles - stall_cycles);
            app->UpdateLLCMissAccessRate(llc_misses, llc_accesses);

            app->UnsetProfiling();
            app->UnsetProfilingNode();

            app->SetRegionProfiled();
        }
    }

#endif
}

void CHeuristicsProfiler::MixedProfiling()
{
#if defined(AMD)
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
    elapsed_time /= 1000; // microseconds

    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    const int TOTAL_CYCLES = 0;
    const int STALL_CYCLES = 1;
    const int LLC_ACCESSES = 2;
    const int LLC_MISSES = 3;
    const int MCT_READS = 4;
    const int MCT_WRITES = 5;
    const int NUM_NODES = NUM_CORES / NUM_CORES_IN_NODE;

    int tiles_per_cluster = arch_->GetCPUNode(0)->num_tiles();

    int num_cpu_nodes = arch_->num_cpu_nodes();
    for (int i = 0; i < num_cpu_nodes; i++) {
        CPUNode* cpu_node = arch_->GetCPUNode(i);

        unsigned long sum_total_cycles = 0;
        unsigned long sum_stall_cycles = 0;

        for (int j = 0; j < cpu_node->num_tiles(); j++) {
            CPUTile* tile = cpu_node->GetTile(j);
            int tile_id = tile->id();

            unsigned long total_cycles = end_.values[TOTAL_CYCLES][tile_id]
                - start_.values[TOTAL_CYCLES][tile_id];
            unsigned long stall_cycles = end_.values[STALL_CYCLES][tile_id]
                - start_.values[STALL_CYCLES][tile_id];

            tile->UpdateTotalCycles_Epoch(total_cycles);
            tile->UpdateStallCycles_Epoch(stall_cycles);

            double cpu_utilization = (double)(total_cycles - stall_cycles/4.0) / (double)max_cycles;
            if (cpu_utilization > 1.0) cpu_utilization = 1.0;
            if (cpu_utilization < 0.1) cpu_utilization = 0.1;
            tile->UpdateCPUUtilization(cpu_utilization);

#ifdef DEBUG_MODE
            std::cout << "tile: " << tile->id()
                << " app: " << scheduler_->GetParApps().find(tile->GetAppID())->second->pid()
                << " total_cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " max_cycles: " << max_cycles
                << " cpu_utilization: " << cpu_utilization << std::endl;
#endif

            sum_total_cycles += total_cycles;
            sum_stall_cycles += stall_cycles;
        }

        cpu_node->UpdateTotalCycles_Epoch(sum_total_cycles);
        cpu_node->UpdateStallCycles_Epoch(sum_stall_cycles);

#ifdef DEBUG_MODE
        std::cout << "MixedProfiling total_cycles: " << sum_total_cycles
            << " stall_cycles: " << sum_stall_cycles
            << " cpu_node: " << i << std::endl;
#endif

        for (int j = 0; j < cpu_node->num_tiles(); j++) {
            CPUTile* tile = cpu_node->GetTile(j);
            int tile_id = tile->id();

            int delegate_tile = tile_id - tile_id % tiles_per_cluster;
            if (tile_id == delegate_tile) {
                unsigned long llc_accesses =
                    end_.values[LLC_ACCESSES][delegate_tile] -
                    start_.values[LLC_ACCESSES][delegate_tile];
                unsigned long llc_misses =
                    end_.values[LLC_MISSES][delegate_tile] -
                    start_.values[LLC_MISSES][delegate_tile];
                unsigned long mct_read_requests =
                    end_.values[MCT_READS][delegate_tile] -
                    start_.values[MCT_READS][delegate_tile];
                unsigned long mct_write_requests =
                    end_.values[MCT_WRITES][delegate_tile] -
                    start_.values[MCT_WRITES][delegate_tile];
                unsigned long mct_requests = mct_read_requests + mct_write_requests;

                cpu_node->UpdateLLCAccesses_Epoch(llc_accesses);
                cpu_node->UpdateLLCMisses_Epoch(llc_misses);
                cpu_node->UpdateDRAMRequests_Epoch(mct_requests);

#ifdef DEBUG_MODE
                std::cout << "MixedProfiling LLC accesses: " << llc_accesses
                    << " LLC misses: " << llc_misses
                    << " MCT reads: " << mct_read_requests
                    << " MCT writes: " << mct_write_requests << std::endl;
#endif
                //cpu_node->UpdateLLCAccessRate_Epoch(llc_accesses, total_cycles);
                //cpu_node->UpdateLLCMissRate_Epoch(llc_misses, total_cycles);
                //cpu_node->UpdateDRAMReqRate_Epoch(dram_requests, total_cycles);
            }
        }
    }
#elif defined(INTEL)
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
    elapsed_time /= 1000; // microseconds

    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    int num_cpu_nodes = arch_->num_cpu_nodes();
    for (int i = 0; i < num_cpu_nodes; i++) {
        CPUNode* cpu_node = arch_->GetCPUNode(i);

        unsigned long sum_total_cycles = 0;
        unsigned long sum_stall_cycles = 0;

        for (int j = 0; j < cpu_node->num_tiles(); j++) {
            CPUTile* tile = cpu_node->GetTile(j);
            int tile_id = tile->id();

            unsigned long total_cycles = getRefCycles(start_.cstate[tile_id], end_.cstate[tile_id]);
            unsigned long stall_cycles = getNumberOfCustomEvents(2, start_.cstate[tile_id], end_.cstate[tile_id]);

            tile->UpdateTotalCycles_Epoch(total_cycles);
            tile->UpdateStallCycles_Epoch(stall_cycles);

            double cpu_utilization = (double)(total_cycles - stall_cycles/4.0) / (double)max_cycles;
            if (cpu_utilization > 1.0) cpu_utilization = 1.0;
            if (cpu_utilization < 0.1) cpu_utilization = 0.1;
            tile->UpdateCPUUtilization(cpu_utilization);

            std::cout << "tile: " << tile->id()
                << " app: " << scheduler_->GetParApps().find(tile->GetAppID())->second->pid()
                << " total_cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " max_cycles: " << max_cycles
                << " cpu_utilization: " << cpu_utilization << std::endl;

            sum_total_cycles += total_cycles;
            sum_stall_cycles += stall_cycles;
        }

        cpu_node->UpdateTotalCycles_Epoch(sum_total_cycles);
        cpu_node->UpdateStallCycles_Epoch(sum_stall_cycles);

        std::cout << "MixedProfiling total_cycles: " << sum_total_cycles
            << " stall_cycles: " << sum_stall_cycles
            << " cpu_node: " << i << std::endl;

        unsigned long llc_accesses = 0;
        unsigned long llc_misses = 0;
        for (int j = 0; j < cpu_node->num_tiles(); j++) {
            CPUTile* tile = cpu_node->GetTile(j);
            int tile_id = tile->id();

            llc_misses += getL3CacheMisses(start_.cstate[tile_id], end_.cstate[tile_id]);
            llc_accesses += getL3CacheHits(start_.cstate[tile_id], end_.cstate[tile_id]) +
                getL3CacheMisses(start_.cstate[tile_id], end_.cstate[tile_id]);
        }

        cpu_node->UpdateLLCAccesses_Epoch(llc_accesses);
        cpu_node->UpdateLLCMisses_Epoch(llc_misses);
        //cpu_node->UpdateDRAMRequests_Epoch(dram_requests);

        std::cout << "MixedProfiling LLC accesses: " << llc_accesses
            << " LLC misses: " << llc_misses << std::endl;
    }

#endif
}

//------------------------------------------------------------------------------
// CC3POProfiler
CC3POProfiler::CC3POProfiler(CPUArchitecture* arch,
        CoreAllocator* scheduler)
    : arch_(arch), scheduler_(scheduler)
{
#if defined(AMD)
    pmu_manager_ = new PMUReqManager();

    int num_events = 5;
    attr_mapping_ = (std::pair<__u32, __u64>*)
        malloc(num_events * sizeof(std::pair<__u32, __u64>));

    attr_mapping_[0] = make_pair<__u32, __u64>
        (TOTAL_CYCLES_TYPE, TOTAL_CYCLES_CONFIG);
    attr_mapping_[1] = make_pair<__u32, __u64>
        (STALL_CYCLES_TYPE, STALL_CYCLES_CONFIG);
    attr_mapping_[2] = make_pair<__u32, __u64>
        (LLC_ACCESSES_TYPE, LLC_ACCESSES_CONFIG);
    attr_mapping_[3] = make_pair<__u32, __u64>
        (LLC_MISSES_TYPE, LLC_MISSES_CONFIG);
    attr_mapping_[4] = make_pair<__u32, __u64>
        (CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_ALL_TYPE,
         CPU_TO_DRAM_REQUESTS_TO_TARGET_NODE_ALL_CONFIG);

    attr_type_ = (event_type*)malloc(num_events * sizeof(event_type));
    attr_type_[0] = PER_CORE;
    attr_type_[1] = PER_CORE;
    attr_type_[2] = PER_NODE;
    attr_type_[3] = PER_NODE;
    attr_type_[4] = PER_NODE;

    pmu_manager_->InitPerfEventAttr(attr_mapping_, attr_type_, num_events);
    pmu_manager_->OpenEventFDs();
#elif defined(INTEL)
    pmu_manager_ = new PMUReqManager();
    pmu_manager_->RecordAllCounters();

#endif

    cout << "Hello! I'm a profiler" << endl;
}

CC3POProfiler::~CC3POProfiler()
{
    cout << "Bye~ from the profiler" << endl;

#if defined(AMD)
    free(attr_mapping_);
    free(attr_type_);
    pmu_manager_->CloseEventFDs();
#elif defined(INTEL)
    delete pmu_manager_;
#endif
}

void CC3POProfiler::ProfilerInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "memory contention profiler info. " << endl;

    out.close();

    return;
}

void CC3POProfiler::StartProfiling()
{
#if defined(AMD)
    pmu_manager_->GetCurrentCount(start_);
#elif defined(INTEL)
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(start_);
#endif

    clock_gettime(CLOCK_REALTIME, &start);
}

void CC3POProfiler::StopProfiling()
{
    struct timespec stop;
    clock_gettime(CLOCK_REALTIME, &stop);

    double elapsed_time = (stop.tv_sec - start.tv_sec)*1000000000 + (stop.tv_nsec - start.tv_nsec);
    elapsed_time /= 1000; // microseconds

    std::cout << "profiling_time: " << (unsigned long)elapsed_time << std::endl;

#if defined(AMD)
    const int TOTAL_CYCLES = 0;
    const int STALL_CYCLES = 1;
    const int LLC_ACCESSES = 2;
    const int LLC_MISSES = 3;
    const int DRAM_REQUESTS = 4;
    const int tiles_per_cluster = arch_->GetCPUNode(0)->num_tiles();

    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    rte_event_values end_;

    unsigned long* per_thread_llc_misses;
    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(end_);

    /* don't save profiling info. if PMU reading request is within 10ms */
    if (end_.values[0][0] == start_.values[0][0]) return;

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();
    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {

        ParApp* app = it->second;

#if DEBUG_MODE
        std::cout << "[Profiler] App (" << app->pid() << ") "
            << "max. cycles: " << max_cycles << std::endl;
#endif

        size_t tiles = app->GetNumTiles();

        list<CPUTile*> tile_list = app->GetTileSet();

        for (list<CPUTile*>::iterator tile_it = tile_list.begin();
                tile_it != tile_list.end();
                tile_it++) {
            CPUTile* tile = *tile_it;
            int tile_id = tile->id();
            unsigned long total_cycles = end_.values[TOTAL_CYCLES][tile_id]
                - start_.values[TOTAL_CYCLES][tile_id];
            unsigned long stall_cycles = 0;
            unsigned long llc_accesses = 0;
            unsigned long llc_misses = 0;
            unsigned long dram_requests = 0;

            stall_cycles = end_.values[STALL_CYCLES][tile_id] -
                start_.values[STALL_CYCLES][tile_id];
            int delegate_tile = tile_id - tile_id % tiles_per_cluster;
            llc_accesses = end_.values[LLC_ACCESSES][delegate_tile] -
                start_.values[LLC_ACCESSES][delegate_tile];
            llc_misses = end_.values[LLC_MISSES][delegate_tile] -
                start_.values[LLC_MISSES][delegate_tile];
            dram_requests = end_.values[DRAM_REQUESTS][delegate_tile] -
                start_.values[DRAM_REQUESTS][delegate_tile];

            //std::cout << "[clean_profiling] app_region: "
            //    << app->GetParallelRegion() << std::endl;

            tile->UpdateTotalCycles_Epoch(total_cycles);
            tile->UpdateStallCycles_Epoch(stall_cycles);

#if DEBUG_MODE
            std::cout << "\tTile " << tile->id()
                << " cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " max_cycles: " << max_cycles << std::endl;
#endif

            double cpu_util = (double)(total_cycles - (stall_cycles/4.0)) / (double)max_cycles;
            if (cpu_util > 1.0)
                cpu_util = 1.0;
            tile->UpdateCPUUtilization(cpu_util);
            tile->UpdateLLCMissRate(((double)llc_misses/(double)tiles_per_cluster)/ max_cycles);
        }
    }

#elif defined(INTEL)
    unsigned long max_cycles = (unsigned long)
       (arch_->max_clock_frequency_MHz()) * elapsed_time;

    rte_event_values end_;

    pmu_manager_->RecordAllCounters();
    pmu_manager_->GetCurrentCount(end_);

    map<pid_t, ParApp*> apps = scheduler_->GetParApps();
    for (map<pid_t, ParApp*>::iterator it = apps.begin();
            it != apps.end();
            it++) {
        ParApp* app = it->second;

#if DEBUG_MODE
        std::cout << "[Profiler] App (" << app->name() << ") "
            << "max. cycles: " << max_cycles << std::endl;
#endif

        list<CPUTile*> tile_list = app->GetTileSet();
        for (list<CPUTile*>::iterator tile_it = tile_list.begin();
                tile_it != tile_list.end();
                tile_it++) {
            CPUTile* tile = *tile_it;
            int tile_id = tile->id();

            unsigned long total_cycles = getRefCycles(start_.cstate[tile_id], end_.cstate[tile_id]);
            unsigned long stall_cycles = getNumberOfCustomEvents(2, start_.cstate[tile_id], end_.cstate[tile_id]);
            // in Intel, LLC misses and DRAM accesses are collected at per-core level
            unsigned long llc_misses = getL3CacheMisses(start_.cstate[tile_id], end_.cstate[tile_id]);

#if DEBUG_MODE
            std::cout << "\tTile " << tile->id()
                << " total_cycles: " << total_cycles
                << " stall_cycles: " << stall_cycles
                << " max_cycles: " << max_cycles
                << " llc_miss_rate: " << (double)llc_misses/(double)max_cycles << std::endl;
#endif

            double cpu_util = (double)(total_cycles - (stall_cycles/4.0)) / (double)max_cycles;
            if (cpu_util > 1.0) cpu_util = 1.0;
            tile->UpdateCPUUtilization(cpu_util);
            tile->UpdateLLCMissRate((double)llc_misses/ max_cycles);
        }
    }

#endif
}
