//------------------------------------------------------------------------------
/// @brief RTE online performance profiler
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
//------------------------------------------------------------------------------

#ifndef __PROFILER_INCLUDED_
#define __PROFILER_INCLUDED_

#include "arch.h"
#include "scheduler.h"
#include "app.h"
#include "rte_cpu.h"
#include "rte_event.h"
#include "pmu_interface.h"

#include <map>
#include <thread>
#include <cstdio>
#include <iostream>

class PMUReqManager;

//------------------------------------------------------------------------------
/// @brief RTE profiler interface
///
class CProfiler {
  public:
    /// @name constructors/destructors
    /// @{

    CProfiler() {}
    ~CProfiler() {}

    /// @}

    /// @name profiler attributes
    /// @{

    virtual void ProfilerInfo(string path) = 0;
    virtual void StartProfiling(void) = 0;
    virtual void StopProfiling(void) = 0;

    /// @}
};

//------------------------------------------------------------------------------
/// @brief RTE online profiler
///
class CSimpleProfiler: public CProfiler {
  public:
    /// @name constructors/destructors
    /// @{

    CSimpleProfiler(CPUArchitecture* arch, CoreAllocator* scheduler);
    ~CSimpleProfiler();

    /// @}

    /// @name profiler attributes
    /// @{

    virtual void ProfilerInfo(string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(void);

    /// @}

  private:
    //
    CPUArchitecture* arch_;         ///< target architecture instance
    CoreAllocator*   scheduler_;    ///< scheduler instance

    PMUReqManager*   pmu_manager_;  ///< pmu request manager

    std::pair<__u32, __u64>* attr_mapping_;
    event_type* attr_type_;
    rte_event_values start_;        ///< pmu counter values at StartProfiling()

    struct timespec start;
};

enum ProfilingPhase {
    NO_PROFILING,      ///< no profiling
    CLEAN_PROFILING,   ///< allocated in CPU node-level
    MIXED_PROFILING,   ///< allocated in mixed
};

class CHeuristicsProfiler: public CProfiler {
  public:
    /// @name constructors/destructors
    /// @{

    CHeuristicsProfiler(CPUArchitecture* arch, CoreAllocator* scheduler);
    ~CHeuristicsProfiler();

    /// @}

    /// @name profiler attributes
    /// @{

    virtual void ProfilerInfo(string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(void);

    void SetProfilingPhase(ProfilingPhase phase) {
        profiling_phase_ = phase;
    }

    ProfilingPhase GetProfilingPhase(void) {
        return profiling_phase_;
    }

    /// @}

  private:
    void CleanProfiling(void);
    void MixedProfiling(void);

    ProfilingPhase profiling_phase_;

    //
    CPUArchitecture* arch_;         ///< target architecture instance
    CoreAllocator*   scheduler_;    ///< scheduler instance

    PMUReqManager*   pmu_manager_;  ///< pmu request manager

    std::pair<__u32, __u64>* attr_mapping_;
    event_type* attr_type_;
    rte_event_values start_;        ///< pmu counter values at StartProfiling()
    rte_event_values end_;

    struct timespec start;
};

//------------------------------------------------------------------------------
/// @brief RTE C3PO online profiler
///
class CC3POProfiler: public CProfiler {
  public:
    /// @name constructors/destructors
    /// @{

    CC3POProfiler(CPUArchitecture* arch, CoreAllocator* scheduler);
    ~CC3POProfiler();

    /// @}

    /// @name profiler attributes
    /// @{

    virtual void ProfilerInfo(string path);
    virtual void StartProfiling(void);
    virtual void StopProfiling(void);

    /// @}

  private:
    //
    CPUArchitecture* arch_;         ///< target architecture instance
    CoreAllocator*   scheduler_;    ///< scheduler instance

    PMUReqManager*   pmu_manager_;  ///< pmu request manager

    std::pair<__u32, __u64>* attr_mapping_;
    event_type* attr_type_;
    rte_event_values start_;        ///< pmu counter values at StartProfiling()

    struct timespec start;
};

#endif
