//------------------------------------------------------------------------------
/// @brief ARM - RTE prototype instance
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
//------------------------------------------------------------------------------

// architecture description
#include "arch.h"

// performance profiler
#include "profiler.h"

// RTE policy specific scheduler
#include "scheduler.h"

// RTE and application connection
#include "app_comm.h"

//------------------------------------------------------------------------------
/// @brief Many-SC prototype
///
/// build basic and flexible prototype structure from architecture description
/// the prototype contains basic RTE functionalities such as policy specific
/// scheduler, performance profiler, architecture specific features and so on.
///
class CPrototype {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief default constructor
    CPrototype();

    /// @brief destructor
    ~CPrototype();

    /// @}

    /// @name prototype attributes
    /// @{

    /// @brief write prototype brief info. into a log file
    ///
    /// @param pointer to architecture instance
    /// @param pointer to scheduler instance
    /// @param pointer to profiler instance
    /// @param path log file path
    void BriefSystemInfo(CPUArchitecture* arch, CoreAllocator* scheduler,
        CProfiler* profiler, string path);

    /// @brief write prototype detailed info. into a log file
    ///
    /// @param pointer to architecture instance
    /// @param pointer to scheduler instance
    /// @param pointer to profiler instance
    /// @param path log file path
    void DetailedSystemInfo(CPUArchitecture* arch, CoreAllocator* scheduler,
        CProfiler* profiler, string path);

    /// @}

  public:
    /// @name prototype attributes
    /// @{

    /// @brief create architecture instance
    static CPUArchitecture*
    CreateArchitectureInstance();

    /// @brief create scheduler instance
    ///
    /// @param pointer to architecture instance
    /// @retval pointer to scheduler instance
    static CoreAllocator*
    CreateScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateEqualPartitionScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateInterleavingScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateMaximizeSpeedupScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateMaximizeUtilizationScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateHeuristicsScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateGreedyScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateC3POScheduler(CPUArchitecture* arch);

    static CoreAllocator*
    CreateLocalityAwareAllocator(CPUArchitecture* arch, int num_cores_per_node);

    static CoreAllocator*
    CreateSimpleAllocator(CPUArchitecture* arch, int num_allocated_cores);

    /// @brief create profiler instance that collects memory access rates
    ///        to figure out applications' scalability
    ///
    /// @param pointer to architecture instance
    /// @param pointer to scheduler instance
    /// @retval pointer to profiler instance
    static CProfiler*
    CreateSimpleProfiler(CPUArchitecture* arch, CoreAllocator* scheduler);

    static CProfiler*
    CreateHeuristicsProfiler(CPUArchitecture* arch,
        CoreAllocator* scheduler);


    /// @brief create profiler instance that collects total CPU cycles
    ///        to figure out applications' CPU utilization
    ///
    /// @param pointer to architecture instance
    /// @param pointer to scheduler instance
    /// @retval pointer to profiler instance
    static CProfiler*
    CreateC3POProfiler(CPUArchitecture* arch, CoreAllocator* scheduler);

    /// @brief create connection between RTE and applications
    ///
    /// @retval application connection instance
    static RTEAppConnection*
    CreateApplicationSocketConnection(void);

    /// @}
};

