//------------------------------------------------------------------------------
/// @brief RTE scheduler
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription scheduler.cpp
/// Basic implementation of the dynamic space-shared scheduler.
/// The scheduler allocates resources (i.e., cores, memories)
/// for each appplication.
/// The allocated resources can not be reclaimed during application lifetime.
///
/// It basically schedule parallel applications through performance profiles
/// in some greedy ways.
/// It will support other optimization techniques and heuristics.
///
//------------------------------------------------------------------------------

#include "scheduler.h"

#include <stdlib.h>
#include <sys/time.h>

static long long getTimestamp(void)
{
    static __thread struct timeval tv;
    gettimeofday(&tv, NULL);
    return 1000000LL * tv.tv_sec + tv.tv_usec;
}

//------------------------------------------------------------------------------
CoreAllocator::CoreAllocator(CPUArchitecture* arch)
    : policy_(EQUAL_PARTITION), arch_(arch), init_c3po_sched(false)
{
   scheduling_phase_ = NO_SCHEDULING;

   num_cores_per_node_ = 0;
   num_cores_ = 0;
}

CoreAllocator::~CoreAllocator()
{
}

/* if the application does not have performance profiles */
ParApp* CoreAllocator::InsertParApp(pid_t pid, string name)
{
    lock();
    static int delegate = 0;

    ParApp* parApp = new ParApp(pid, name, this, arch_);
    if (!parApp) {
        cout << "application insertion failed" << endl;
    } else {
        // add to the list: pid is the key of map list
        apps_.insert(pair<pid_t, ParApp*>(parApp->pid(), parApp));
    }


    parApp->SetDelegate(delegate);
    delegated_list_.push_back(delegate);
    cout << "insert delegate " << delegate << endl;
    delegate++;

    init_c3po_sched = false;
    unlock();

    return parApp;
}

ParApp* CoreAllocator::FindParAppByPID(pid_t pid)
{
    map<pid_t, ParApp*>::iterator it = apps_.find(pid);

    if (!apps_.size()) {
        cout << "couldn't find the application (pid: " << pid << ")" << endl;
        return NULL;
    }

    return it->second;
}

void CoreAllocator::DeleteParApp(pid_t pid)
{
    lock();
    map<pid_t, ParApp*>::iterator it = apps_.find(pid);

    if (!apps_.size()) {
        cout << "couldn't find the application (pid: " << pid << ")" << endl;

        unlock();
        return;
    }

    // deallocate all allocated tiles
    ParApp* app = it->second;
    app->FlushAllTiles();
    app->ReleaseSHM();

    int delegate = app->GetDelegate();
    std::list<int>::iterator d_it;
    for (std::list<int>::iterator i_it = delegated_list_.begin();
            i_it != delegated_list_.end();
            i_it++) {
        int item = *i_it;
        if (delegate == item) {
            d_it = i_it;
            break;
        }
    }

    delegated_list_.erase(d_it);
    cout << "delete delegate " << delegate << endl;

    apps_.erase(pid);

    delete app;

    init_c3po_sched = false;
    unlock();

    return;
}

void CoreAllocator::Schedule()
{
    lock();
    switch (GetPolicy()) {
        case EQUAL_PARTITION:
        {
            ReclaimCoreResource();
            EqualPartitionScheduler();
            break;
        }
        case INTERLEAVING:
        {
            ReclaimCoreResource();
            InterleavingScheduler();
            break;
        }
        case MAX_SPEEDUP:
        {
            ReclaimCoreResource();
            MaximizeSpeedupScheduler();
            break;
        }
        case MAX_UTILIZATION:
        {
            ReclaimCoreResource();
            MaximizeUtilizationScheduler();
            break;
        }
        case HEURISTICS:
        {
            HeuristicsScheduler();
            break;
        }
        case GREEDY:
        {
            GreedyScheduler();
            break;
        }
        case LOCALITY_AWARE:
        {
            LocalityAwareScheduler();
            break;
        }
        case SIMPLE:
        {
            SimpleScheduler();
            break;
        }
        case CLEAN_CORE_ALLOCATION:
        {
            ReclaimCoreResource();
            break;
        }
        case C3PO:
        {
            C3POScheduler();
            break;
        }
        default:
        {
            cerr << "[scheduler][ERROR] not such scheduling policy." << endl;
          // XXX should we abort?
        }
    }
    unlock();
}

void CoreAllocator::ReclaimCoreResource(void)
{
    for (std::map<pid_t, ParApp*>::iterator it = apps_.begin();
            it != apps_.end();
            it++) {
        ParApp* app = it->second;
        app->FlushAllTiles();
    }
}

void CoreAllocator::ShowAppsInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "Application Information" << endl;

    for (std::map<pid_t, ParApp*>::iterator it = apps_.begin();
            it != apps_.end();
            it++) {
        ParApp* app = it->second;
        app->ShowAppInfo(path);
    }

    out << endl;

    out.close();

    return;
}

void CoreAllocator::SchedulerInfo(string path)
{
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "detailed scheduler information" << endl;
    out << "policy: " << GetPolicy() << endl;

    out.close();

    return;
}

void CoreAllocator::ShowScheduleResult(string path)
{
    long long time = getTimestamp();
    ofstream out;

    out.open(path.c_str(), ios_base::out | ios_base::app);

    out << "========================================"
        << "========================================" << endl;
    out << "Application Scheduling Result" << endl;
    out << "sec: " << time/1000000LL << "." << time%1000000LL << endl;

    for (std::map<pid_t, ParApp*>::iterator it = apps_.begin();
            it != apps_.end();
            it++) {
        ParApp* app = it->second;
        app->ShowAppResource(out);
    }

    out.close();
}

