//------------------------------------------------------------------------------
/// @brief RTE scheduler
/// @author Younghyun Cho <younghyun@csap.snu.ac.kr>
/// Computer Systems and Platforms Laboratory
/// Department of Computer Science and Engineering
/// Seoul National University
///
/// @section license_section Licence
/// Copyright (c) 2016 Computer Systems and Platforms Laboratory
/// All rights reserved.
///
/// @decription scheduler.h
/// Basic description of the dynamic space-shared scheduler.
/// The scheduler allocates resources (i.e., cores, memories)
/// for each appplication.
/// The allocated resources can not be reclaimed during application lifetime.
///
/// It basically schedule parallel applications through performance profiles
/// in some greedy ways.
/// It will support other optimization techniques and heuristics.
///
//------------------------------------------------------------------------------

#ifndef __SCHEDULER_INCLUDED_
#define __SCHEDULER_INCLUDED_

#include "common.h"
#include "arch.h"
#include "app.h"
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <string>

#include <list>
#include <mutex>

#include "profiler.h"
#include "perf_model.h"

using namespace std;

class CProfiler;
class CHeuristicsProfiler;

//------------------------------------------------------------------------------
/// @brief RTE resource allocation policies
///
/// each member of this enumeration represents a policy in the RTE
///
enum Policy {
    EQUAL_PARTITION,  ///< equal partition
    INTERLEAVING,     ///< interleaving partition
    MAX_PERFORMANCE,  ///< maximize the throughput
    C3PO,             ///< C3PO core allocation
    MAX_SPEEDUP,      ///< maximize speedup
    MAX_UTILIZATION,  ///< maximize CPU and memory utilization
    HEURISTICS,
    GREEDY,
    LOCALITY_AWARE,
    SIMPLE,
    CLEAN_CORE_ALLOCATION,
};

enum SchedulingPhase {
    NO_SCHEDULING,
    PROFILING_RUN,
    GREEDY_ALLOCATION,
    ADAPTIVE_ALLOCATION,
};

//------------------------------------------------------------------------------
/// @brief default static scheduler
///
/// static scheduler assume time 0 scheduling
/// it does policy specific scheduling on target architecture
///
class CoreAllocator {
  public:
    /// @name constructors/destructors
    /// @{

    /// @brief constructor taking initialization values
    CoreAllocator(CPUArchitecture* arch);

    /// @brief destructor
    ~CoreAllocator();

    /// @}

    /// @name scheduler attributes
    /// @{

    std::list<int> GetDelegatedList(void) { return delegated_list_; }

    /// @brief set new policy
    void SetPolicy(Policy policy) { policy_ = policy; }

    void SetNumCoresPerNode(int num_cores_per_node)
    { num_cores_per_node_ = num_cores_per_node; }
    void SetNumCores(int num_cores)
    { num_cores_ = num_cores; }

    /// @brief get current policy
    Policy GetPolicy(void) const { return policy_; }

    /// @brief get the number of currently running apps
    int RunningApps(void) const { return apps_.size(); }

    /// @brief insert an application to scheduling list
    ///
    /// @param pid     application id
    /// @param name    application name
    ParApp* InsertParApp(pid_t pid, string name);

    /// @brief get the list of applications running on the scheduler
    map<pid_t, ParApp*> GetParApps(void) { return apps_; }

    /// @brief get application pointer of given pid
    ParApp* FindParAppByPID(pid_t pid);

    /// @brief delete an application from scheduling list
    void DeleteParApp(pid_t pid);

    /// @brief scheduler RTE policy specific scheduling
    void Schedule();

    /// @brief show scheduling results
    void ShowScheduleResult(string path);

    /// @brief write all applications info. into a log file
    void ShowAppsInfo(string path);

    /// @brief write scheduler info. into a log file
    void SchedulerInfo(string path);

    void RegisterProfiler(CProfiler* profiler) {
        profiler_ = profiler;
    }

    void SetSchedulingPhase(SchedulingPhase phase) {
        scheduling_phase_ = phase;
    }

    SchedulingPhase GetSchedulingPhase(void) {
        return scheduling_phase_;
    }

    void lock(void) {
        scheduler_lock_.lock();
    }

    void unlock(void) {
        scheduler_lock_.unlock();
    }

    /// @}

  private:
    /// @name scheduler private attributes
    /// @{

    /// @brief system-wide resource reclamation
    void ReclaimCoreResource(void);

    /// @brief clear cluster map
    ///
    /// @description re-initialize cluster map for new scheduling
    void ClearClusterMap(void);

    /// @}


    /// @name scheduling algorithms
    /// @{

    bool EqualPartitionScheduler(void);
    bool InterleavingScheduler(void);
    bool C3POScheduler(void);
    bool MaximizeSpeedupScheduler(void);
    bool MaximizeUtilizationScheduler(void);
    bool LocalityAwareScheduler(void);
    bool SimpleScheduler(void);
    bool HeuristicsScheduler(void);
    bool GreedyScheduler(void);
    bool BaseMapping(int last_cpu_node);
    bool BaseEQP(void);
    bool Scheduler_Phase_1(void);
    bool Scheduler_Phase_1_Adaptive(void);
    bool Scheduler_Phase_1_Clean(void);
    bool Scheduler_Phase_2(void);
    bool Scheduler_Phase_2_MaxSpeedup(void);
    bool Scheduler_Phase_3_CPU_Utilization(void);
    bool Scheduler_Phase_3_LLC_Efficiency(void);

    bool D_AdaptiveScheduler(void);
    void AdaptiveClusterScheduling(int clusters);
    void AdaptiveTileScheduling(int cl);

    /// @}

    std::mutex scheduler_lock_;

    SchedulingPhase scheduling_phase_;

    CProfiler* profiler_;

    map<pid_t, ParApp*> apps_; ///< application list

    Policy policy_;            ///< RTE policy

    CPUArchitecture* arch_;    ///< the target architecture description

    /// @name C3PO variables
    /// @{

    bool init_c3po_sched;     ///< C3PO (run once) for new/del app
    vector<CPUTile*> deallocated_tiles;   ///< set of deallocated tiles' ids

    /// @}

    std::list<int> delegated_list_;

    int num_cores_per_node_;
    int num_cores_;
};

#endif // __SCHEDULER_INCLUDED_

